from tradssat.tmpl.var import CharacterVar, FloatVar, IntegerVar

vars_ = {
    CharacterVar('INSI', 4, spc=2, info='Institute + Site code'),
    FloatVar('LAT', 8, 3, info='Latitude, degrees (decimals)'),
    FloatVar('LONG', 8, 3, info='Longitude, degrees (decimals)'),
    FloatVar('ELEV', 5, 0, info='Elevation, m'),
    FloatVar('TAV', 5, 1, info='Air temperature average, °C'),
    FloatVar('AMP', 5, 1, info='Air temperature amplitude, monthly averages, °C'),
    FloatVar('REFHT', 5, 1, info='Height of temperature measurements, m'),
    FloatVar('WNDHT', 5, 1, info='Height of wind measurements, m'),
    IntegerVar('CCO2', 4, info='Atmospheric CO2, ppm'),

    CharacterVar('DATE', 5, spc=0, info='Year + days from Jan. 1'),
    FloatVar('SRAD', 5, 1, info='Solar radiation, MJ m-2 day-1'),
    FloatVar('TMAX', 5, 1, info='Air temperature maximum, °C'),
    FloatVar('TMIN', 5, 1, info='Air temperature minimum, °C'),
    FloatVar('RAIN', 5, 1, info='Precipitation, mm'),
    FloatVar('DEWP', 5, 1, info='Dewpoint temperature5, °C'),
    FloatVar('WIND', 5, 1, info='Wind run, km day-1'),
    FloatVar('PAR', 5, 1, info='Photosynthetic active radiation (PAR)5, moles m-2 day-1'),
    FloatVar('EVAP', 5, 1),
    FloatVar('RHUM', 5, 1)

}



from tradssat.tmpl.var import FloatVar, IntegerVar

vars_ = {
    IntegerVar('StYr', 5, info='Start year'),
    IntegerVar('StMn', 5, info='Start month'),
    IntegerVar('SpYr', 5, info='Stop year'),
    IntegerVar('SpMn', 5, info='Stop month'),

    IntegerVar('yr', 4, spc=0, info='Year'),
    IntegerVar('mo', 2, info='Month'),
    FloatVar('srmn', 5, 2, info='Solar radiation mean, MJ m-2 day-1'),
    FloatVar('srsd', 5, 2, info='Solar radiation standard deviation, MJ m-2 day-1'),
    FloatVar('txmn', 5, 2, info='Temperature maximum mean, °C'),
    FloatVar('txsd', 5, 2, info='Temperature maximum standard deviation, °C'),
    FloatVar('tnmn', 5, 2, info='Temperature minimum mean, °C'),
    FloatVar('tnsd', 5, 2, info='Temperature minimum standard deviation, °C'),
    FloatVar('ramn', 5, 2, info='Rainfall mean, mm'),
    FloatVar('rasd', 5, 2, info='Rainfall standard deviation, mm'),

}



from tradssat import ExpFile
from tradssat.exper.exper_vars import TRT_HEAD
from .mgr import PeriphFileMgr


class ExpFileMgr(PeriphFileMgr):

    def __init__(self, file):
        self.file = ExpFile(file)

    def get_file_val(self, var, subsect=None, sect=None):
        return self.file.get_value(var, subsect=subsect, sect=sect)

    def set_file_val(self, var, val, subsect=None, sect=None, cond=None):
        self.file.set_value(var, val, sect=sect, subsect=subsect, cond=cond)

    def get_trt_nums(self):
        return self.file.get_value('N', sect=TRT_HEAD)

    def get_trt_names(self):
        return self.file.get_value('TNAME', sect=TRT_HEAD)

    def add_row(self, sect, subsect=None, vals=None):
        self.file.add_row(sect=sect, subsect=subsect, vals=vals)

    def remove_row(self, sect, subsect=None, cond=None):
        self.file.remove_row(sect=sect, subsect=subsect, cond=cond)

    def find_var_sect(self, var):
        return self.file.find_var_sect(var)

    def get_value(self, var, level):
        sect = self.file.find_var_sect(var)
        lv_cd = _level_codes[_factor_to_code[sect]]

        return self.file.get_value(var, sect=sect, cond={lv_cd: level})

    def set_value(self, var, val, level):
        sect = self.file.find_var_sect(var)
        lv_cd = _level_codes[_factor_to_code[sect]]

        self.file.set_value(var, val, sect=sect, cond={lv_cd: level})

    def variables(self):
        return self.file.variables()


_level_codes = {
    'CU': 'C',
    'FL': 'L',
    'SA': 'A',
    'IC': 'C',
    'MP': 'P',
    'MI': 'I',
    'MF': 'F',
    'MR': 'R',
    'MC': 'C',
    'MT': 'T',
    'ME': 'E',
    'MH': 'H',
    'SM': 'N'
}
_factor_codes = {
    'CU': 'CULTIVARS',
    'FL': 'FIELDS',
    'SA': 'SOIL ANALYSIS',
    'IC': 'INITIAL CONDITIONS',
    'MP': 'PLANTING DETAILS',
    'MI': 'IRRIGATION AND WATER MANAGEMENT',
    'MF': 'FERTILIZERS (INORGANIC)',
    'MR': 'RESIDUES AND ORGANIC FERTILIZER',
    'MC': 'CHEMICAL APPLICATIONS',
    'MT': 'TILLAGE AND ROTATIONS',
    'ME': 'ENVIRONMENT MODIFICATIONS',
    'MH': 'HARVEST DETAILS',
    'SM': 'SIMULATION CONTROLS'
}
_factor_to_code = {fct: cd for cd, fct in _factor_codes.items()}



from tradssat.tmpl.var import CharacterVar, FloatVar, IntegerVar

cul_vars_CSCAS = {
    CharacterVar('VAR#', 6, spc=0, info='Identification code or number for the specific cultivar.'),
    CharacterVar('VAR-NAME', 16, header_fill='.', info='Name of cultivar.'),
    CharacterVar('EXP#', 5, miss='.', info='Number of experiments used for calibration.'),
    CharacterVar('ECO#', 6, info='Ecotype code for this cultivar,points to entry in the ECO file'),

    FloatVar('PPS1', 5, 0, info='Photoperiod sensitivity for phase n. (% drop for 10h pp.change)'),

    FloatVar('B01ND', 5, 2, info='Duration from branch 0 to branch 1 (ie.tier x,node number)'),
    FloatVar('B12ND', 5, 2, info='Duration from branch 1 to branch 2 (ie.tier x,node number)'),
    FloatVar('B23ND', 5, 2, info='Duration from branch 2 to branch 3 (ie.tier x,node number)'),
    FloatVar('B34ND', 5, 2, info='Duration from branch 3 to branch 4 (ie.tier x,node number)'),
    FloatVar('B45ND', 5, 2, info='Duration from branch 4 to branch 5 (ie.tier x,node number)'),
    FloatVar('B56ND', 5, 2, info='Duration from branch 5 to branch 6 (ie.tier x,node number)'),

    FloatVar('SR#WT', 5, 2, info='Storage root number per unit canopy weight at initiation (#/g)'),
    FloatVar('SRFR', 5, 2, info='Fr.of assimilate designated for tops sent to storage root (#)'),
    FloatVar('HMPC', 5, 0, info='Harvest product moisture content (%)'),
    FloatVar('PHINT', 5, 0, info='Interval between leaf tip appearances for first leaves (oC.d)'),
    FloatVar('LA1S', 5, 0, info='Area/leaf (cm2) of the first leaves when growing without stress.'),
    FloatVar('LAXS', 5, 0, info='Area/leaf at maximum area/leaf (cm2)'),
    IntegerVar('LAXND', 5, info='Node # at which maximum potential area/leaf reached (#)'),
    IntegerVar('LAXN2', 5, info='Node # at which potential area/leaf begins to decline (#)'),
    FloatVar('LAFS', 5, 0, info='End of cycle area/leaf (cm2)'),
    IntegerVar('LAFND', 5, info='Node # at which the end of cycle area/leaf reached (#)'),
    IntegerVar('SLAS', 5, info='Specific leaf lamina area when crop growing without stress (cm2/g)'),
    FloatVar('LLIFA', 5, 0, info='Leaf life,from full expansion to start senescence (Thermal units)'),
    FloatVar('LPEFR', 5, 2, info='Leaf petiole fraction (fr of lamina+petiole)'),
    FloatVar('STFR', 5, 2, info='Stem fraction of assimilate destined for canopy growth (fr)'),

}

eco_vars_CSCAS = {
    CharacterVar('ECO#', 6, spc=0, info='Code for the ecotype to which a cultivar belongs (see *.cul file)'),
    CharacterVar('ECONAME', 17, header_fill='.', info='Name of the ecotype, which is referenced from *.CUL file'),

    FloatVar('BR1FX', 5, 0, info='Branch number per fork at fork 1, maximum (#)'),
    FloatVar('BR2FX', 5, 0, info='Branch number per fork at fork 2, maximum (#)'),
    FloatVar('BR3FX', 5, 0, info='Branch number per fork at fork 3, maximum (#)'),
    FloatVar('BR4FX', 5, 0, info='Branch number per fork at fork 4, maximum (#)'),
    FloatVar('BR5FX', 5, 0, info='Branch number per fork at fork 5, maximum (#)'),
    FloatVar('BR6FX', 5, 0, info='Branch number per fork at fork 6, maximum (#)'),
    FloatVar('PARUE', 5, 2, info='PAR conversion factor,standard (g dry matter/MJ)'),
    FloatVar('DUSRI', 5, 0, info='Degree days at which storage root number determined (oC.d)'),
    FloatVar('SRN%S', 5, 2, info='Storage root standard N concentration (% dm)'),
    FloatVar('HTSTD', 5, 0, info='Standard canopy height (cm)')
}



from tradssat.tmpl.output import OutFile
from tradssat.tmpl.var import FloatVar, IntegerVar


class PlantGroOut(OutFile):
    """
    File reader for DSSAT plant growth (PLANTGRO.OUT) output files.
    """
    filename = 'PlantGro.OUT'

    def _get_var_info(self):
        return vars_

vars_ = {
    IntegerVar('TREATMENT', 2, info='Treatment number'),

    IntegerVar('YEAR', 4, info='Year'),
    IntegerVar('DOY', 3, info='Day of year starting on Jan 1.'),
    IntegerVar('DAS', 3, info='Day after start'),
    IntegerVar('DAP', 3, info='Days after planting'),
    FloatVar('L#SD', 4, 0, info='Leaf number'),
    FloatVar('GSTD', 4, 4, info='Growth stage'),
    FloatVar('LAID', 4, 2, info='Leaf Area Index'),
    IntegerVar('CWAD', 4, info='Tops dry weight, kg/Ha'),
    IntegerVar('VWAD', 4, info='Veg dry weight, kg/Ha'),
    IntegerVar('LWAD', 4, info='Leaf dry weight, kg/Ha'),
    IntegerVar('SWAD', 4, info='Stem dry weight, kg/Ha'),
    IntegerVar('FLWAD', 5, info='Flower dry weight, kg/Ha'),
    IntegerVar('FWAD', 4, info='Fruit dry weight, kg/Ha'),
    IntegerVar('CRAD', 4, info='Crown dry weight, kg/Ha'),
    IntegerVar('BWAD', 4, info='Basal dry weight, kg/Ha'),
    IntegerVar('SUGD', 4, info='Suck dry weight, kg/Ha'),
    IntegerVar('RWAD', 4, info='Root dry weight, kg/Ha'),
    FloatVar('HIAD', 4, 3, info='Harvest index'),
    IntegerVar('EYWAD', 5, info='Eye Weight, kg/Ha'),
    IntegerVar('EY#AD', 5, info='Eye number'),
    FloatVar('WSPD', 4, 3, info='Water stress in photosynthesis'),
    FloatVar('WSGD', 4, 3, info='Water stress in growth'),
    FloatVar('NSTD', 4, 3, info='Nitrogen stress'),
    FloatVar('LN%D', 4, 2, info='Leaf Nitrogen percentage'),
    FloatVar('SLAD', 4, 1, info='Specific Leaf area'),
    FloatVar('RDPD', 4, 1, info='Root depth, m'),
    FloatVar('RL1D', 4, 2, info='Level 1 Root Length density, cm3/cm3 of soil'),
    FloatVar('RL2D', 4, 2, info='Level 2 Root Length density, cm3/cm3 of soil'),
    FloatVar('RL3D', 4, 2, info='Level 3 Root Length density, cm3/cm3 of soil'),
    FloatVar('RL4D', 4, 2, info='Level 4 Root Length density, cm3/cm3 of soil'),
    FloatVar('RL5D', 4, 2, info='Level 5 Root Length density, cm3/cm3 of soil'),
    FloatVar('RL6D', 4, 2, info='Level 6 Root Length density, cm3/cm3 of soil'),
    FloatVar('RL7D', 4, 2, info='Level 7 Root Length density, cm3/cm3 of soil'),
    FloatVar('RL8D', 4, 2, info='Level 8 Root Length density, cm3/cm3 of soil'),
    FloatVar('RL9D', 4, 2, info='Level 9 Root Length density, cm3/cm3 of soil'),
    FloatVar('RL10', 4, 2, info='Level 10 Root Length density, cm3/cm3 of soil'),

    # Recently added (#19)
    FloatVar('TMEAN', 5, 1, info='Mean Temperature (TMAX + TMIN/2, ºC'),
    FloatVar('PARID', 5, 3, info='PAR interception (%)'),
    FloatVar('PARUD', 5, 2, info='PAR utilization efficiency (g/MJ)'),
    FloatVar('AWAD', 4, 1, info='Assimilate Production (kg/(ha.d))'),
    FloatVar('SAID', 4, 3, info='Stem Area Index (m2/m2)'),
    FloatVar('CAID', 4, 4, info='Canopy Area Index'),
    IntegerVar('TWAD', 4, info='Tops + Roots + storage wt (kg[dm]/ha)'),
    IntegerVar('SDWAD', 5, info='Seed weight (kg/ha)'),
    IntegerVar('HWAD', 4, info='Harvest Product wt (kg [dm]/ha)'),
    IntegerVar('RSWAD', 5, info='Reserves weight (kg/ha)'),
    FloatVar('SNWLD', 5, 3, info='Senesced OM added to surface (kg/ha)'),
    FloatVar('SNWSD', 5, 4, info='Senesced OM added to soil (kg/ha)'),
    FloatVar('RS%D', 4, 2, info='Reserves Concentration (%)'),
    IntegerVar('S#AD', 4, info='Shoot (apex) Number (no/m2)'),
    FloatVar('SWXD', 4, 1, info='Extractable water (mm)'),
    FloatVar('WAVRD', 5, 1, info='Water available to demand ratio (#)'),
    FloatVar('WUPRD', 5, 2, info='Water uptake to demand ratio (#)'),
    FloatVar('WFPD', 4, 2, info='Water factor for photosynthesis (0-1)'),
    FloatVar('WFGD', 4, 2, info='Water factor for growth (0-1)'),
    FloatVar('NFPD', 4, 2, info='N factor for photosynthesis (0-1)'),
    FloatVar('NFGD', 4, 2, info='N factor for leaf growth (0-1)'),
    FloatVar('NUPRD', 5, 1, info='N uptake to demand ratio (#)'),
    FloatVar('TFPD', 4, 2, info='Temperature factor for photosyntesis (0-1)'),
    FloatVar('TFGD', 4, 2, info='Temperature factor for leaf growth (0-1)'),
    FloatVar('DYLFD', 5, 2, info='Daylength factor for development (0-1)'),

    IntegerVar('GWAD', 4, info='Grain dry weight (kg [dm]/ha)'),
    IntegerVar('G#AD', 4, info='Grain number (no/m2)'),
    FloatVar('GWGD', 4, 1, info='Unit grain weight (mg [dm]/grain'),
    IntegerVar('PWAD', 4, info='Pod weight (kg [dm]/ha)'),
    IntegerVar('P#AD', 4, info='Pod number (no/m2)'),
    FloatVar('EWSD', 4, 3, info='Excess water stress (0-1)'),
    FloatVar('PST1A', 5, 3, info='P stress factor for reducing photosynthate (0-1)'),
    FloatVar('PST2A', 5, 3, info='P stress factor which affects vegetative growth (0-1)'),
    FloatVar('KSTD', 4, 3, info='Potassium stress factor (0-1)'),
    FloatVar('SH%D', 4, 2, info='Shelling % (seed weight/pod wt * 100)'),
    FloatVar('HIPD', 4, 2, info='Pod harvest index (pod/top)'),
    IntegerVar('PWDD', 4, info='Detached pod mass (kg [dm]/ha)'),
    IntegerVar('PWTD', 4, info='Total pod weight (kg [dm]/ha)'),
    FloatVar('CHTD', 4, 2, info='Canopy height (m)'),
    FloatVar('CWID', 4, 2, info='Canopy width (m, for 1 row)'),
    FloatVar('RDPD', 4, 2, info='Root depth (m)'),
    IntegerVar('CDAD', 4, info='Dead canopy weight (kg [dm]/ha)'),
    IntegerVar('LDAD', 4, info='Dead leaf weight (kg [dm]/ha)'),
    IntegerVar('SDAD', 4, info='Dead stem weight (kg [dm]/ha)'), 
    IntegerVar('SNW0C', 5, info='Cumulative senesced dry matter to surface (kg/ha)'),
    IntegerVar('SNW1C', 5, info='Cumulative senesced dry matter to soil (kg/ha)'),
    FloatVar('DTTD', 4, 2, info='Growing degree days (ºC -d/d)'),
    FloatVar('NWAD', 4, 1, info='Nodule weight (kg [dm]/ha)'),
    FloatVar('TKILL', 4,1, info='Temperature for plant death (ºC)'),
    IntegerVar('CHWAD', 5, info='Chaff weight (kg/ha)'),
    IntegerVar('EWAD', 4, info='Ear (grain + chaff) or eye weight (kg [dm]/ha)'),
    IntegerVar('RSWAD', 5, info='Reserves weight (kg/ha)'),
    IntegerVar('SNWPD', 5, info='Dead material retained on plant (kg/ha)'),
    FloatVar('SNWLD', 5, 2, info='Senesced OM added to surface (kg/ha)'),
    FloatVar('SNWSD', 5, 4, info='Senesced OM added to soil (kg/ha)'),
    IntegerVar('H#AD', 4, info='Number (no/m2)'),
    FloatVar('HWUD', 4, 4, info='Harvest weight (wt/unit)'),
    IntegerVar('T#AD', 4, info='Tiller numer (no/m2)'),
    FloatVar('PTFD', 4, 2, info='Partition fraction of assimilates to shoot (0-1)'),
    FloatVar('WFTD', 4, 2, info='Water factor for transpiration (0-1)'),
    FloatVar('NFTD', 4, 2, info='N factor for tillering (0-1)'),
    FloatVar('VRNFD', 5, 2, info='Vernalization factor (0-1)')

}




from tradssat.genotype.vars_._cropsim import cropsim_cul_vars, cropsim_eco_vars

cul_vars_WHCRP = cropsim_cul_vars()

eco_vars_WHCRP = cropsim_eco_vars(exclude=['GWTAF', 'GWTAT', 'G#RF', 'RTNUP', 'NUPNF', 'NUPWF'])



import os

from .utils import set_dssat_dir, get_dssat_dir, read_txt
from .exper import ExpFile
from .genotype import CULFile, ECOFile
from .out import PlantGroOut, SummaryOut, SoilNiOut, SoilTempOut, SoilWatOut, ETOut
from .soil import SoilFile
from .weather import WTHFile, MTHFile, CLIFile
from .mgrs import DSSATRun, DSSATResults, GeneticMgr, WeatherFileMgr, SoilMgr

try:
    set_dssat_dir('C:/DSSAT47')
except FileNotFoundError:
    pass



from tradssat.genotype.vars_._cropgro import cropgro_cul_vars, cropgro_eco_vars
from tradssat.tmpl.var import FloatVar

cul_vars_BMGRO = cropgro_cul_vars()

eco_vars_BMGRO = cropgro_eco_vars()
eco_vars_BMGRO.update({
    FloatVar('RDRMT', 5, dec=3,
             info='Relative dormancy sensitivity of this cultivar to daylength - partitioning (0-1)'),
    FloatVar('RDRMG', 5, dec=3,
             info='Relative dormancy sensitivity of this cultivar to daylength - photosynthesis (0-1)'),
    FloatVar('RDRMM', 5, dec=3,
             info='Relative dormancy sensitivity of this cultivar to daylength - mobilization (0-1)'),
    FloatVar('RCHDP', 5, dec=3, info='Relative cold hardening potential (0-1)')
})



import os

from tradssat import WTHFile
from .mgr import PeriphFileMgr, get_dssat_subdir


class PeriphWeatherMgr(PeriphFileMgr):

    def __init__(self, codes, levels):
        self.files = {lvl: WeatherFileMgr(cd) for cd, lvl in zip(codes, levels)}

    def get_value(self, var, level):
        self.files[level].get_value(var)

    def set_value(self, var, val, level):
        self.files[level].set_value(var, val)

    def variables(self):
        return {str(vr) for f in self.files.values() for vr in f.variables()}

    def write(self, force=False):
        for f in self.files.values():
            f.write(force)


class WeatherFileMgr(object):

    def __init__(self, code):
        weather_dir = get_dssat_subdir('Weather')
        gen_weather_dir = os.path.join(weather_dir, 'Gen')

        self.file = None
        for d in [weather_dir, gen_weather_dir]:
            for f in os.listdir(d):
                if WTHFile.matches_file(f):
                    name = os.path.split(os.path.splitext(f)[0])[1]
                    if name.startswith(code):
                        self.file = WTHFile(os.path.join(d, f))
                        break
        if self.file is None:
            raise ValueError('No weather file found matching "{}".'.format(code))

    def get_value(self, var):
        return self.file.get_value(var)

    def variables(self):
        return self.file.variables()



import re


def _name_matches(pattern, name, full=False):
    """
    
    Parameters
    ----------
    pattern: str | re.Pattern
        The desired name, or a regex.
    name: str
        The name to test.
    full:
        Whether ``pattern`` should match the full name or just the beginning.

    Returns
    -------

    """
    if None in [pattern, name]:
        return True

    elif isinstance(pattern, str):

        if (pattern == name) or (not full and name.startswith(pattern)):
            return pattern

    else:
        m = re.fullmatch(pattern, name) if full else re.match(pattern, name)

        if m:
            return m.group()



from tradssat.tmpl.output import OutFile
from tradssat.tmpl.var import FloatVar, IntegerVar


class ETOut(OutFile):
    filename = 'ET.OUT'

    def _get_var_info(self):
        return vars_


vars_ = {
    IntegerVar('YEAR', 4, info='Year'),
    IntegerVar('DOY', 3, info='Day of year starting on Jan 1.'),
    IntegerVar('DAS', 3, info='Day after start'),
    FloatVar('SRAA', 4, 2, info='Average solar radiation (MJ/(m2.day'),
    FloatVar('TMAXA', 5, 2, info='Average maximum air temperature (°C)'),
    FloatVar('TMINA', 5, 2, info='Average minimum air temperature (°C)'),
    FloatVar('EOAA', 4, 3, info='Average potential evapotranspiration (mm/d)'),
    FloatVar('EOPA', 4, 3, info=' Average potential transpiration (mm/d)'),
    FloatVar('EOSA', 4, 3, info='Average potential soil evaporation per day (mm/day)'),
    FloatVar('ETAA', 4, 3, info='Average evapotranspiration (mm/d)'),
    FloatVar('EPAA', 4, 0, info=' Average plant transpiration (mm/d)'),
    FloatVar('ESAA', 4, 3, info='Average soil evaporation (mm/d)'),
    FloatVar('EFAA', 4, 0, info='Average flood evaporation (mm/d)'),
    FloatVar('EMAA', 4, 3, info='Average mulch evaporation (mm/d)'),
    FloatVar('EOAC', 4, 2, info='Cumulative potential evapotranspiration (mm)'),
    FloatVar('ETAC', 4, 2, info='Cumulative evapotranspiration (mm)'),
    FloatVar('EPAC', 4, 0, info=' Cumulative transpiration (mm)'),
    FloatVar('ESAC', 4, 2, info='Cumulative soil evaporation (mm)'),
    FloatVar('EFAC', 4, 0, info='Cumulative floodwater evaporation (mm)'),
    FloatVar('EMAC', 4, 2, info='Cumulative mulch evaporation (mm)'),

    # Recently added (#19)
    FloatVar('REFA', 4, 3, info='ASCE standardized reference ET'),
    FloatVar('KCAA', 4, 3, info='FAO-56 single crop coefficient'),
    FloatVar('KCBA', 4, 3, info='FAO-56 basal crop coefficient'),
    FloatVar('KEAA', 4, 3, info='FAO-56 evaporation coefficient'),
    FloatVar('ES1D', 4, 3, info='Soil evaporation Layer 1 (mm/d)'),
    FloatVar('ES2D', 4, 3, info='Soil evaporation Layer 2 (mm/d)'),
    FloatVar('ES3D', 4, 3, info='Soil evaporation Layer 3 (mm/d)'),
    FloatVar('ES4D', 4, 3, info='Soil evaporation Layer 4 (mm/d)'),
    FloatVar('ES5D', 4, 3, info='Soil evaporation Layer 5 (mm/d)'),
    FloatVar('ES6D', 4, 3, info='Soil evaporation Layer 6 (mm/d)'),
    FloatVar('ES7D', 4, 3, info='Soil evaporation Layer 7 (mm/d)'),
    FloatVar('ES8D', 4, 3, info='Soil evaporation Layer 8 (mm/d)'),
    FloatVar('ES9D', 4, 3, info='Soil evaporation Layer 9 (mm/d)'),
    FloatVar('ES10D', 5, 3, info='Soil evaporation Layer 10 (mm/d)'),
    FloatVar('TRWUD', 5, 3, info='Root water uptake (cm/d)'),
}



from tradssat.tmpl.var import CharacterVar, FloatVar

cul_vars_RICER = {
    CharacterVar('VAR#', 6, spc=0, info='Identification code or number for the specific cultivar.'),
    CharacterVar('VAR-NAME', 16, header_fill='.', info='Name of cultivar.'),
    CharacterVar('EXPNO', 5, miss='.', info='Number of experiments used for calibration.'),
    CharacterVar('ECO#', 6, info='Code for the ecotype to which this cultivar belongs (currently not used)'),

    FloatVar(
        'P1 ', 5, 1,
        info='Time period (expressed as growing degree days [GDD] in øC above a base temperature of 9øC) from seedling '
             'emergence during which the rice plant is not responsive to changes in photoperiod. This period is also '
             'referred to as the basic vegetative phase of the plant.'
    ),
    FloatVar(
        'P2O', 5, 1,
        info='Critical photoperiod or the longest day length (in hours) at which the development occurs at a maximum '
             'rate. At values higher than P2O developmental rate is slowed, hence there is delay due to longer day '
             'lengths.'
    ),
    FloatVar(
        'P2R', 5, 1,
        info='Extent to which phasic development leading to panicle initiation is delayed (expressed as GDD in øC) '
             'for each hour increase in photoperiod above P2O.'
    ),
    FloatVar(
        'P5 ', 5, 1,
        info='Time period in GDD øC) from beginning of grain filling (3 to 4 days after flowering) to physiological '
             'maturity with a base temperature of 9øC.'
    ),
    FloatVar(
        'G1 ', 5, 1,
        info='Potential spikelet number coefficient as estimated from the number of spikelets per g of main culm dry '
             'weight (less lead blades and sheaths plus spikes) at anthesis. A typical value is 55.'
    ),
    FloatVar(
        'G2 ', 5, 4,
        info='Single grain weight (g) under ideal growing conditions, i.e. nonlimiting light, water, nutrients, '
             'and absence of pests and diseases.'
    ),
    FloatVar(
        'G3 ', 5, 2,
        info='Tillering coefficient (scaler value) relative to IR64 cultivar under ideal conditions. A higher '
             'tillering cultivar would have coefficient greater than 1.0.'
    ),
    FloatVar(
        'G4 ', 5, 2,
        info='Temperature tolerance coefficient. Usually 1.0 for varieties grown in normal environments. G4 for '
             'japonica type rice growing in a warmer environment would be 1.0 or greater. Likewise, the G4 value for '
             'indica type rice in very cool environments or season would be less than 1.0.'
    ),
    FloatVar('PHINT', 5, 1, info=''),
    FloatVar('G5', 5, 1, info='')

}



import os

from tradssat import SoilFile
from .mgr import get_dssat_subdir


class PeriphSoilMgr(object):

    def __init__(self, soils, levels):
        self.files = {lvls: SoilMgr(sl) for sl, lvls in zip(soils, levels)}

    def get_value(self, var, level):
        return self.files[level].get_value(var)

    def set_value(self, var, val, level):
        return self.files[level].set_value(var, val)

    def variables(self):
        return list({v for f in self.files.values() for v in f.variables()})


class SoilMgr(object):
    def __init__(self, code):

        soils_dir = get_dssat_subdir('Soil')

        self.file = None
        for f in os.listdir(soils_dir):
            if SoilFile.matches_file(f):
                file = SoilFile(os.path.join(soils_dir, f))

                if code in file:
                    self.file = file
                    self.code = code
                    break

        if self.file is None:
            raise ValueError('No soil file found for soil "{}".'.format(code))

    def get_value(self, var):
        return self.file.get_value(var, sect=self.code)

    def set_value(self, var, val):
        return self.file.set_value(var, val, sect=self.code)

    def variables(self):
        return self.file.variables()



import numpy as np

from tradssat.exper.exper_vars import TRT_HEAD, GENERAL
from tradssat.mgrs.exp_mgr import _level_codes, _factor_codes, _factor_to_code
from .exp_mgr import ExpFileMgr
from .gen_mgr import PeriphGenMgr
from .soil_mgr import PeriphSoilMgr
from .wth_mgr import PeriphWeatherMgr


def _valid_factor(factor):
    if factor in _factor_codes:
        return factor
    else:
        try:
            return _factor_to_code[factor]
        except KeyError:
            raise ValueError('Factor "{}" not recognised.'.format(factor))


class DSSATRun(object):
    """
    General manager for DSSAT run input files.
    """

    def __init__(self, file, model=None):
        """
        
        Parameters
        ----------
        file: str
            Input experiment file (.ccX).
        model: str
            Crop model code (optional). Useful when a crop can be simulated by more than one DSSAT crop model.
        """
        self.file = file
        self.exp = ExpFileMgr(file)

        field_levels = self.exp.get_file_val('L', sect='FIELDS')
        self.soil = PeriphSoilMgr(self.exp.get_file_val('ID_SOIL'), field_levels)
        self.weather = PeriphWeatherMgr(self.exp.get_file_val('WSTA'), levels=field_levels)

        cult_levels = self.exp.get_file_val('C', sect='CULTIVARS')
        crops = self.exp.get_file_val('CR')
        cults = self.exp.get_file_val('INGENO')
        self.genetics = PeriphGenMgr(crops, cults, cult_levels, model)

        self.peripherals = [self.soil, self.weather, self.genetics]

        self.check()

    def get_general_val(self, var):
        """
        Obtain a variable value from the `GENERAL` section of the EXP file.
        
        Parameters
        ----------
        var: str
            Variable name

        Returns
        -------
        np.ndarray
            Variable value.
        """
        return self.exp.get_file_val(var, sect=GENERAL)

    def set_general_val(self, var, val):
        """
        Sets a variable value in the `GENERAL` section of the EXP file.

        Parameters
        ----------
        var: str
            Variable name.
        val: str | float | int | np.ndarray
            New value for the variable.
        """
        self.exp.set_file_val(var, val, sect=GENERAL)

    def add_treatment(self, name, ops=None, factors=None):
        """
        Adds a treatment level to the experiment.
        
        Parameters
        ----------
        name: str
            Name of the new treatment.
        ops: dict
            Dictionnary of values for the `R`, `O`, and `C` treatment level options (optional).
        factors: dict
            Dictionnary of values for the treatment's factor levels (optional). Missing factors will be assigned 
            level 0.
        """
        if factors is None:
            factors = {}
        if ops is None:
            ops = {}

        default_ops = {'R': 1, 'O': 0, 'C': 0}
        default_ops.update(ops)

        factors = {_valid_factor(cd): vl for cd, vl in factors.items()}

        n = self.treatments().size + 1
        d_vals = {
            'N': n,
            'TNAME': name,
            **default_ops,
            **{cd: 0 for cd in _factor_codes}
        }
        d_vals.update(factors)

        self.exp.add_row(sect=TRT_HEAD, subsect=0, vals=d_vals)

    def remove_treatment(self, trt):
        """
        Removes a treatment from the experiment.
        
        Parameters
        ----------
        trt: str | int
            Treatment name or number.
        """
        if isinstance(trt, str):
            trt = self.get_trt_num(trt)
        self.exp.remove_row(sect=TRT_HEAD, subsect=0, cond={'N': trt})

    def add_factor_level(self, factor, vals):
        """
        Adds a factor level to the experiment.
        
        Parameters
        ----------
        factor: str
            Factor code or name.
        vals: dict
            Dictionnary of variable values for that factor level's variables. Missing variables will be assigned
            the corresponding missing code (usually -99).

        """
        factor = _valid_factor(factor)
        factor_name = _factor_codes[factor]
        level = self.n_factor_levels(factor) + 1
        lv_code = _level_codes[factor]
        level_vals = {
            lv_code: level,
            **vals
        }
        self.exp.add_row(sect=factor_name, vals=level_vals)

    def get_trt_factor_level(self, trt, factor):
        """
        Obtain the factor level for a specific treatment.

        Parameters
        ----------
        trt: str | int
            Treatment name or number.
        factor: str
            Factor name or code.

        Returns
        -------
        int
            The factor level corresponding to the given treatment.
        """
        factor = _valid_factor(factor)
        trt = self._valid_trt(trt)

        trt_nums = self.treatments()
        levels = self.exp.get_file_val(factor, sect=TRT_HEAD)

        return levels[trt_nums == trt][0]

    def set_trt_factor_level(self, trt, factor, level):
        """
        Sets the factor level for a treatment.
        
        Parameters
        ----------
        trt: str | int
            The treatment name or number.
        factor: str
            The factor name or code.
        level: int
            The new factor level.
        """
        factor = _valid_factor(factor)
        trt = self._valid_trt(trt)

        self.exp.set_file_val(factor, level, sect=TRT_HEAD, subsect=0, cond={'N': trt})

    def get_factor_level_val(self, var, level):
        """
        Obtain the variable value for a specific factor level.
        
        Parameters
        ----------
        var: str
            The variable of interest.
        level: int
            The factor level corresponding to the specified variable.

        Returns
        -------
        np.ndarray
            The variable value at the factor level.
        """
        file = self._locate_var(var)
        return file.get_value(var, level=level)

    def set_factor_level_val(self, var, val, level):
        file = self._locate_var(var)
        file.set_value(var, val, level=level)

    def get_trt_val(self, var, trt):
        """
        Returns the value of a treatment's variable.
        
        Parameters
        ----------
        var: str
            The variable of interest.
        trt: int | str
            The treatment name or number.

        Returns
        -------
        np.ndarray
            The variable value for the treatment.
        """
        var_loc = self._locate_var(var)
        factor = self._get_var_factor(var, var_loc)
        level = self.get_trt_factor_level(trt, factor)

        return var_loc.get_value(var, level)

    def set_trt_val(self, var, val, trt):

        var_loc = self._locate_var(var)
        factor = self._get_var_factor(var, var_loc)
        level = self.get_trt_factor_level(trt, factor)

        var_loc.set_value(var, val, level)

    def treatments(self, name=False):
        """
        Returns the treatments in the run.
        
        Parameters
        ----------
        name: bool
            Whether to return treatment names or numbers.

        Returns
        -------
        np.ndarray
            The list of treatments.
        """
        
        if name:
            return self.exp.get_trt_names()
        else:
            return self.exp.get_trt_nums()

    def n_factor_levels(self, factor):
        factor = _valid_factor(factor)
        factor_name = _factor_codes[factor]
        lv_var = _level_codes[factor]

        levels = self.exp.get_file_val(lv_var, sect=factor_name)

        return np.unique(levels).size

    def get_trt_name(self, n):
        """
        Returns the treatment name corresponding to a treatment number.
        
        Parameters
        ----------
        n: int
            The treatment number.

        Returns
        -------
        str
            The treatment name.
        """
        nums = self.treatments()
        names = self.treatments(name=True)

        return names[np.where(nums == n)][0]

    def get_trt_num(self, trt):
        """
        Returns the treatment number corresponding to a specified treatment name.
        
        Parameters
        ----------
        trt: str
            The treatment name.

        Returns
        -------
        int
            The corresponding treatment number.
        """
        nums = self.treatments()
        names = self.treatments(name=True)

        return nums[np.where(names == trt)]

    def check(self):
        # check consistent treatment numbering (numeric order)
        # Check all treatments exist
        # check no ununsed treatments (warn)
        # check all external soil, weather, genetic files exist
        # Check all treatment factor levels exist in sections
        pass

    def clean(self):
        # remove factor levels not used in any treatment

        # renumber treatments
        trts = np.arange(self.treatments().size)
        self.exp.set_file_val('N', trts, sect=TRT_HEAD)

        # renumber factor levels, and reorder them in subsections
        pass

    def write(self, file=None, force=False, clean=True, override=False):
        if clean:
            self.clean()

        for prph in self.peripherals:
            prph.write(force=force)

        self.exp.file.write(file=file, force=force)

    def _valid_trt(self, trt):

        if trt is None:
            return self.treatments()

        elif isinstance(trt, int):
            trt = [self.get_trt_name(trt)]
        elif isinstance(trt, str):
            trt = [trt]

        for i, t in enumerate(trt):
            if isinstance(t, str):
                trt[i] = self.get_trt_num(t)[0]

            elif t not in self.treatments():
                raise ValueError('Treatment "{t}" not found.'.format(t=t))

        return trt

    def _locate_var(self, var):

        for f in [self.exp, self.soil, self.genetics, self.weather]:
            if var in f.variables():
                return f

        raise ValueError('Could not find variable "{}".'.format(var))

    def _get_var_factor(self, var, var_loc):
        if isinstance(var_loc, (PeriphSoilMgr, PeriphWeatherMgr)):
            sect = 'FIELDS'
        elif isinstance(var_loc, PeriphGenMgr):
            sect = 'CULTIVARS'
        else:
            sect = self.exp.find_var_sect(var)

        return _factor_to_code[sect]



from tradssat.tmpl.var import CharacterVar, FloatVar

cul_vars_MLCER = {
    CharacterVar('VAR#', 6, spc=0, info='Identification code or number for the specific cultivar.'),
    CharacterVar('VAR-NAME', 16, header_fill='.', info='Name of cultivar.'),
    CharacterVar('EXPNO', 5, miss='.', info='Number of experiments used for calibration.'),
    CharacterVar('ECO#', 6, info='Code for the ecotype to which this cultivar belongs (see *.eco file)'),

    FloatVar(
        'P1', 5, 1,
        info='Thermal time from seedling emergence to the end of the juvenile phase (expressed in degree days above a'
             ' base temperature of 10øC) during which the plant is not responsive to changes in photoperiod.'
    ),
    FloatVar(
        'P20', 5, 2,
        info='Critical photoperiod or the longest day length (in hours) at which development occurs at a maximum rate. '
             'At values greater than P2O, the rate of development is reduced.'
    ),
    FloatVar(
        'P2R', 5, 1,
        info='Extent to which phasic development leading to panicle initiation (expressed in degree days) is delayed '
             'for each hour increase in photoperiod above P2O.'
    ),
    FloatVar(
        'P5', 5, 1,
        info='Thermal time (degree days above a base temperature of 10øC) from beginning of grain filling (3-4 days '
             'after flowering) to physiological maturity.'
    ),
    FloatVar('G1', 5, 2, info='Scaler for relative leaf size.'),
    FloatVar('G4', 5, 2, info='Scaler for partitioning of assimilates to the panicle (head).'),
    FloatVar(
        'PHINT', 5, 2,
        info='Phylochron interval; the interval in thermal time (degree days) between successive leaf tip appearances.'
             ''),
    FloatVar('GT', 5, 2, info='Tillering coefficient, equivalent to G1, but on tillers'),
    FloatVar('G5', 5, 1, info='Potential grain size, mg')

}

eco_vars_MLCER = {
    CharacterVar('ECO#', 6, spc=0, info='Code for the ecotype to which a cultivar belongs (see *.cul file)'),
    CharacterVar('ECONAME', 17, header_fill='.', info='Name of the ecotype, which is referenced from *.CUL file'),

    FloatVar('TBASE', 5, 1, info='base temperature below which no development occurs, C'),
    FloatVar('TOPT', 5, 1, info='temperature at which maximum development rate occurs during vegetative stages, C'),
    FloatVar('ROPT', 5, 1, info='temperature at which maximum development rate occurs for reproductive stages, C'),
    FloatVar(
        'DJTI', 5, 1,
        info='Minimum days from end of juvenile stage to tassel initiation if the cultivar is not photoperiod '
             'sensitive, days'
    ),
    FloatVar('GDDE', 5, 1, info='Growing degree days per cm seed depth required for emergence, GDD/cm'),
    FloatVar('RUE', 5, 1, info='Radiation use efficiency, g plant dry matter/MJ PAR'),
    FloatVar('KCAN', 5, 2, info='Canopy light extinction coefficient for daily PAR.')
}



from tradssat.tmpl.var import CharacterVar, FloatVar, IntegerVar

vars_ = {
    CharacterVar('INSI', 4, spc=2, info='Institute + Site code'),
    FloatVar('LAT', 8, 3, info='Latitude, degrees (decimals)'),
    FloatVar('LONG', 8, 3, info='Longitude, degrees (decimals)'),
    FloatVar('ELEV', 5, 0, info='Elevation, m'),
    FloatVar('TAV', 5, 1, info='Air temperature average, °C'),
    FloatVar('AMP', 5, 1, info='Air temperature amplitude, monthly averages, °C'),
    FloatVar('SRAY', 5, 1, info=''),
    FloatVar('TMXY', 5, 1, info=''),
    FloatVar('TMNY', 5, 1, info=''),
    IntegerVar('RAIY', 5, info=''),

    IntegerVar('START', 5, spc=0, info=''),
    IntegerVar('DURN', 5, info=''),
    FloatVar('ANGA', 5, 2, info=''),
    FloatVar('ANGB', 5, 2, info=''),
    FloatVar('REFHT', 5, 1, info=''),
    FloatVar('WNDHT', 5, 1, info=''),
    CharacterVar('SOURCE', 50, info=''),
    IntegerVar('GSST', 5, info=''),
    IntegerVar('GSDU', 5),

}



from tradssat.tmpl.var import CharacterVar, FloatVar

cul_vars_TNARO = {
    CharacterVar('VAR#', 6, spc=0, info='Identification code or number for the specific cultivar.'),
    CharacterVar('VAR-NAME', 16, header_fill='.', info='Name of cultivar.'),
    CharacterVar('EXPNO', 5, miss='.', info='Number of experiments used to estimate cultivar parameters.'),
    CharacterVar(
        'ECO#', 6,
        info='Ecotype code or this cultivar, points to the Ecotype in the ECO file (currently not used).'
    ),
    FloatVar('P1', 5, 0),
    FloatVar('P3', 5, 0),
    FloatVar('P4', 5, 0),
    FloatVar('P5', 5, 0),
    FloatVar('G2', 5, 2),
    FloatVar('G3', 5, 2),
    FloatVar('G4', 5, 2),
    FloatVar('PHINT', 5, 1),
    FloatVar('PCINT', 5, 1),
    FloatVar('PCGRD', 5, 2)

}



from tradssat.tmpl.var import CharacterVar, FloatVar

cul_vars_PTSUB = {
    CharacterVar('VAR#', 6, spc=0, info='Identification code or number for the specific cultivar.'),
    CharacterVar('VAR-NAME', 16, header_fill='.', info='Name of cultivar.'),
    CharacterVar('EXPNO', 5, miss='.', info='Number of experiments used for calibration.'),
    CharacterVar(
        'ECO#', 6, info='Ecotype code or this cultivar, points to the Ecotype in the ECO file (currently not used).'
    ),

    FloatVar('G2', 5, 0, info='Leaf area expansion rate after tuber initiation (cm2/m2 d)'),
    FloatVar('G3', 5, 1, info='Potential tuber growth rate (g/m2 d)'),
    FloatVar('PD', 5, 1,
             info='Index that supresses tuber growth during the period that immediately follows tuber induction'),
    FloatVar('P2', 5, 1, info='Tuber initiation sensitivity to long photoperiods'),
    FloatVar('TC', 5, 1, info='Upper critical temperature for tuber initiation (C)'),

}

eco_vars_PTSUB = {
    CharacterVar('ECO#', 6, spc=0, info='Code for the ecotype to which a cultivar belongs (see *.cul file)'),
    CharacterVar('ECONAME', 18, header_fill='.', info='Name of the ecotype, which is referenced from *.CUL file'),

    FloatVar('RUE1', 5, 1, info='Radiation use efficiency, ISTAGE=1, g plant dry matter/MJ PAR'),
    FloatVar('RUE2', 5, 1, info='Radiation use efficiency, ISTAGE>1, g plant dry matter/MJ PAR')

}



def _gen_vars(dict_vars):
    vars_set = set()
    for name, d_var in dict_vars.items():
        vars_set.add(d_var['class'](name, **d_var['args']))

    return vars_set


def _return_vars(info_dict, rename=None, exclude=None):
    if exclude is None:
        exclude = []
    if rename is None:
        rename = {}

    info_dict_copy = info_dict.copy()
    for old, new in rename.items():
        info_dict_copy[new] = info_dict_copy.pop(old)

    d_vars = {vr: d_vr for vr, d_vr in info_dict_copy.items() if vr not in exclude}
    return _gen_vars(d_vars)



from tradssat.tmpl.var import CharacterVar, FloatVar
from ._utils import _return_vars

_standard_cul_vars = {
    'VAR#': {
        'class': CharacterVar,
        'args': dict(size=6, spc=0, info='Identification code or number for the specific cultivar.')
    },
    'VRNAME': {'class': CharacterVar, 'args': dict(size=16, header_fill='.', info='Name of cultivar.')},
    'EXPNO': {
        'class': CharacterVar, 'args': dict(size=5, miss='.', info='Number of experiments used for calibration.')
    },  # Yes, I know. Number of experiments is a character var.
    'ECO#': {'class': CharacterVar,
             'args': dict(size=6, info='Code for the ecotype to which this cultivar belongs (see *.eco file)')},
    'CSDL': {'class': FloatVar,
             'args': dict(
                 size=5, dec=2,
                 info='Critical Short Day Length below which reproductive development progresses with no daylength '
                      'effect (for shortday plants) (hour)')
             },
    'PPSEN': {'class': FloatVar,
              'args': dict(
                  size=5, dec=4,
                  info='Slope of the relative response of development to photoperiod with time '
                       '(positive for shortday plants) (1/hour)')
              },
    'EM-FL': {'class': FloatVar,
              'args': dict(size=5, dec=2,
                           info='Time between plant emergence and flower appearance (R1) (photothermal days)')
              },
    'FL-SH': {'class': FloatVar,
              'args': dict(size=5, dec=2, info='Time between first flower and first pod (R3) (photothermal days)')},
    'FL-SD': {'class': FloatVar,
              'args': dict(size=5, dec=2, info='Time between first flower and first seed (R5) (photothermal days)')},
    'SD-PM': {'class': FloatVar,
              'args': dict(size=5, dec=2,
                           info='Time between first seed (R5) and physiological maturity (R7) (photothermal days)')
              },
    'FL-LF': {'class': FloatVar,
              'args': dict(size=5, dec=2,
                           info='Time between first flower (R1) and end of leaf expansion (photothermal days)')},
    'LFMAX': {'class': FloatVar,
              'args': dict(size=5, dec=3,
                           info='Maximum leaf photosynthesis rate at 30 C, 350 vpm CO2, and high light (mg CO2/m2-s)')},
    'SLAVR': {'class': FloatVar,
              'args': dict(size=5, dec=1,
                           info='Specific leaf area of cultivar under standard growth conditions (cm2/g)')},
    'SIZLF': {'class': FloatVar,
              'args': dict(size=5, dec=2, info='Maximum size of full leaf (three leaflets) (cm2)')},
    'XFRT': {'class': FloatVar,
             'args': dict(size=5, dec=2, info='Maximum fraction of daily growth that is partitioned to seed + shell')},
    'WTPSD': {'class': FloatVar,
              'args': dict(size=5, dec=4, info='Maximum weight per seed (g)')},
    'SFDUR': {'class': FloatVar,
              'args': dict(
                  size=5, dec=2,
                  info='Seed filling duration for pod cohort at standard growth conditions (photothermal days)')},
    'SDPDV': {'class': FloatVar,
              'args': dict(size=5, dec=2, info='Average seed per pod under standard growing conditions (#/pod)')},
    'PODUR': {'class': FloatVar,
              'args': dict(
                  size=5, dec=1,
                  info='Time required for cultivar to reach final pod load under optimal conditions (photothermal days)'
              )},
    'THRSH': {'class': FloatVar,
              'args': dict(size=5, dec=1,
                           info='The maximum ratio of (seed/(seed+shell)) at maturity. '
                                'Causes seed to stop growing as their dry weights '
                                'increase until shells are filled in a cohort. '
                                '(Threshing percentage).'
                           )},
    'SDPRO': {'class': FloatVar, 'args': dict(size=5, dec=3, info='Fraction protein in seeds (g(protein)/g(seed))')},
    'SDLIP': {'class': FloatVar, 'args': dict(size=5, dec=3, info='Fraction oil in seeds (g(oil)/g(seed))')}
}

_standard_eco_vars = {
    'ECO#': {'class': CharacterVar,
             'args': dict(size=6, spc=0, info='Code for the ecotype to which a cultivar belongs (see *.cul')
             },
    'ECONAME': {'class': CharacterVar,

                'args': dict(size=17, header_fill='.',
                             info='Name of the ecotype, which is referenced from *.CUL file')},
    'MG': {'class': CharacterVar,
           'args': dict(size=2, info='Maturity group number for this ecotype, such as maturity group in soybean')},
    'TM': {'class': CharacterVar, 'args': dict(size=2, info='Indicator of temperature adaptation')},
    'THVAR': {'class': FloatVar,
              'args': dict(size=5, dec=1,
                           info='Minimum rate of reproductive development under short days and optimal temperature')},
    'PP-SS': {'class': FloatVar, 'args': dict(size=5, dec=2, info='')},
    'PL-EM': {'class': FloatVar,
              'args': dict(size=5, dec=2, info='Time between planting and emergence (V0) (thermal days)')},
    'EM-V1': {'class': FloatVar,
              'args': dict(
                  size=5, dec=1,
                  info='Time required from emergence to first true leaf (V1), thermal days')},
    'V1-JU': {'class': FloatVar,
              'args': dict(size=5, dec=1,
                           info='Time required from first true leaf to end of juvenile phase, thermal days')},
    'JU-R0': {'class': FloatVar,
              'args': dict(
                  size=5, dec=1,
                  info='Time required for floral induction, equal to the minimum number of days for floral induction '
                       'under optimal temperature and daylengths, photothermal days')},
    'PM06': {'class': FloatVar,
             'args': dict(
                 size=5, dec=2,
                 info='Proportion of time between first flower and first pod for first peg (peanut only)')},
    'PM09': {'class': FloatVar,
             'args': dict(
                 size=5, dec=2,
                 info='Proportion of time between first seed and physiological maturity that the last seed can be '
                      'formed')},
    'LNGSH': {'class': FloatVar,
              'args': dict(
                  size=5, dec=1,
                  info='Time required for growth of individual shells (photothermal days)')},
    'R7-R8': {'class': FloatVar,
              'args': dict(
                  size=5, dec=2,
                  info='Time between physiological (R7) and harvest maturity (R8) (days)')},
    'FL-VS': {'class': FloatVar,
              'args': dict(
                  size=5, dec=2,
                  info='Time from first flower to last leaf on main stem (photothermal days)')},
    'TRIFL': {'class': FloatVar,
              'args': dict(
                  size=5, dec=2,
                  info='Rate of appearance of leaves on the mainstem (leaves per thermal day)')},
    'RWDTH': {'class': FloatVar,
              'args': dict(
                  size=5, dec=2,
                  info='Relative width of this ecotype in comparison to the standard width per node (YVSWH) defined '
                       'in the species file (*.SPE)')},
    'RHGHT': {'class': FloatVar,
              'args': dict(
                  size=5, dec=2,
                  info='Relative height of this ecotype in comparison to the standard height per node (YVSHT) defined '
                       'in the species file (*.SPE)')},
    'SIZELF': {'class': FloatVar,
               'args': dict(
                   size=5, dec=2,
                   info='The size of a normal upper node leaf (nodes 8 - 10) of variety I, used to adjust leaf area '
                        'expansion during sink-limited phase of vegetative growth, i.e., prior to VSSINK nodes on the '
                        'main stem (cm2/leaf)')},
    'R1PPO': {'class': FloatVar,
              'args': dict(
                  size=5, dec=3,
                  info='Increase in daylength sensitivity after R1 (CSDVAR and CLDVAR both decrease with the same '
                       'amount) (h)')},
    'OPTBI': {'class': FloatVar,
              'args': dict(
                  size=5, dec=1,
                  info='Minimum daily temperature above which there is no effect on slowing normal development toward '
                       'flowering (oC)')},
    'SLOBI': {'class': FloatVar,
              'args': dict(
                  size=5, dec=3,
                  info='Slope of relationship reducing progress toward flowering if TMIN for the day is less than '
                       'OPTBI')}
}

_std_frm_eco_vars = {
    'RDRMT': {
        'class': FloatVar,
        'args': dict(size=5, dec=3,
                     info='Relative dormancy sensitivity of this cultivar to daylength - partitioning (0-1)')
    },
    'RDRMG': {
        'class': FloatVar,
        'args': dict(size=5, dec=3,
                     info='Relative dormancy sensitivity of this cultivar to daylength - photosynthesis (0-1)')
    },
    'RDRMM': {
        'class': FloatVar,
        'args': dict(size=5, dec=3,
                     info='Relative dormancy sensitivity of this cultivar to daylength - mobilization (0-1)')
    },
    'RCHDP': {'class': FloatVar, 'args': dict(size=5, dec=3, info='Relative cold hardening potential (0-1)')},
}
_std_frm_eco_vars.update(_standard_eco_vars)


def cropgro_cul_vars(var_name='VRNAME', exclude=None):
    if var_name != 'VRNAME':
        rename = {'VRNAME': var_name}
    else:
        rename = None

    return _return_vars(_standard_cul_vars, rename=rename, exclude=exclude)


def cropgro_eco_vars(rename=None, exclude=None):
    return _return_vars(_standard_eco_vars, rename=rename, exclude=exclude)


def frm_eco_vars(rename=None, exclude=None):
    return _return_vars(_std_frm_eco_vars, rename=rename, exclude=exclude)



from tradssat.tmpl import InpFile
from .cli_vars import vars_


class CLIFile(InpFile):
    """
    Reader for DSSAT climate (.CLI) files.
    """

    ext = ['.CLI']

    def _get_var_info(self):
        return vars_



from tradssat.tmpl.output import OutFile
from tradssat.tmpl.var import FloatVar, IntegerVar


class SoilNiOut(OutFile):
    """
    Reader for DSSAT soil nitrogen (SOILNI.OUT) files.
    """
    filename = 'SoilNi.Out'

    def _get_var_info(self):
        return vars_


vars_ = {
    IntegerVar('YEAR', 4, info='Year'),
    IntegerVar('DOY', 3, info='Day of year starting on Jan 1.'),
    IntegerVar('DAS', 3, info='Day after start'),
    IntegerVar('NAPC', 4, info='Cumulative inorganic N applied, kg/ha'),
    IntegerVar('NI#M', 4, info='N application numbers'),
    FloatVar('NIAD', 4, 1, info='Inorganic N in soil, kg/ha'),
    FloatVar('NITD', 4, 1, info='Amount of total NO3, kg/ha'),
    FloatVar('NHTD', 4, 1, info='Amount of total NH4, kg/ha'),
    FloatVar('NI1D', 4, 2, info='NO3 at 0-5 cm soil depth, ppm'),
    FloatVar('NI2D', 4, 2, info='NO3 at 5-15 cm soil depth, ppm'),
    FloatVar('NI3D', 4, 2, info='NO3 at 15-30 cm soil depth, ppm'),
    FloatVar('NI4D', 4, 2, info='NO3 at 30-45 cm soil depth, ppm'),
    FloatVar('NI5D', 4, 2, info='NO3 at 45-60 cm soil depth, ppm'),
    FloatVar('NI6D', 4, 2, info='NO3 at 60-90 cm soil depth, ppm'),
    FloatVar('NI7D', 4, 2, info='NO3 at 90-110 cm soil depth, ppm'),
    FloatVar('NH1D', 4, 2, info='NH4 at 0-5 cm soil depth, ppm'),
    FloatVar('NH2D', 4, 2, info='NH4 at 5-15 cm soil depth, ppm'),
    FloatVar('NH3D', 4, 2, info='NH4 at 15-30 cm soil depth, ppm'),
    FloatVar('NH4D', 4, 2, info='NH4 at 30-45 cm soil depth, ppm'),
    FloatVar('NH5D', 4, 2, info='NH4 at 45-60 cm soil depth, ppm'),
    FloatVar('NH6D', 4, 2, info='NH4 at 60-90 cm soil depth, ppm'),
    FloatVar('NH7D', 4, 2, info='NH4 at 90-110 cm soil depth, ppm'),
    FloatVar('NMNC', 4, 0, info=''),
    FloatVar('NITC', 4, 0, info=''),
    FloatVar('NDNC', 4, 0, info=''),
    FloatVar('NIMC', 4, 0, info=''),
    FloatVar('AMLC', 4, 0, info=''),
    FloatVar('NNMNC', 5, 0, info=''),
    FloatVar('NUCM', 4, 0, info='N uptake, kg/ha'),
    FloatVar('NLCC', 4, 0, info='Cumulative N leached, kg/ha'),
}



from tradssat.tmpl.var import CharacterVar, FloatVar

cul_vars_WHAPS = {
    CharacterVar('VAR#', 6, spc=0, info='Identification code or number for the specific cultivar.'),
    CharacterVar('VRNAME', 16, header_fill='.', info='Name of cultivar.'),
    CharacterVar('EXPNO', 5, miss='.', info='Number of experiments used for calibration.'),
    CharacterVar('ECO#', 6, info='Code for the ecotype to which this cultivar belongs (see *.eco file)'),

    FloatVar('VSEN', 5, 2, info='sensitivity to vernalisation'),
    FloatVar('PPSEN', 5, 2, info='sensitivity to photoperiod'),
    FloatVar('P1', 5, 1, info='Thermal time from seedling emergence to the end of the juvenile phase'),
    FloatVar('P5', 5, 1, info='Thermal time (base 0oC) from beginning of grainfill to maturity: range 500 to 700'),
    FloatVar('PHINT', 5, 1, info='Phyllochron interval'),
    FloatVar(
        'GRNO ', 5, 2,
        info='Coefficient of kernel number per stem weight at the beginning of grain filling [kernels (g stem)-1]'
    ),
    FloatVar('MXFIL', 5, 2, info='Potential kernel growth rate [mg kernel-1 day-1]:  Values between 1.0 and 3.0'),
    FloatVar(
        'STMMX', 5, 2,
        info='Potential final dry weight of a single tiller (excluding grain) (g stem-1) values 1.0 - 3.0'
    ),
    FloatVar('SLAP1', 5, 1, info='ratio of leaf area to mass at emergence (cm2/g)'),
    FloatVar('SLAP2', 5, 1, info='ratio of leaf area to mass at end of leaf growth (cm2/g)'),
    FloatVar('TC1P1', 5, 2, info='for calculating tc1: tiller number from emerg. to term. spik.(=stem elongation)'),
    FloatVar('TC1P2', 5, 2, info='tc1 = tc1_p1 + tc1_p2 *cumph_nw(istage)'),
    FloatVar('DTNP1', 5, 3, info='for calculating dtiln: tiller number after term. spik.(=stem elongation)'),
    FloatVar('PLGP1', 5, 0, info='for calculating plag: potential leaf growth.  plag= plag_p1*cumph(istage)**plag_p2'),
    FloatVar('PLGP2', 5, 2, info='for calculating plag: potential leaf growth.  plag= plag_p1*cumph(istage)**plag_p2'),
    FloatVar('P2AF', 5, 2, info='threshold AD in a layer becoming effective on root growth'),
    FloatVar('P3AF', 5, 1, info='length of downwards root not effected under aeration deficit'),
    FloatVar('P4AF', 5, 2, info='days to be accumulated before aeration deficit effects root growth'),
    FloatVar('P5AF', 5, 2, info='power term at af1'),
    FloatVar('P6AF', 5, 2, info='days to be accumulated before aeration deficit effects crop growth'),
    FloatVar('ADLAI', 5, 2,
             info='threshold aeration deficit (AF2) affecting LAI        (set to 1.0 for no stress run)'),
    FloatVar('ADTIL', 5, 2,
             info='threshold aeration deficit (AF2) effecting tillering  (set to 1.0 for no stress run)'),
    FloatVar('ADPHO', 5, 2,
             info='threshold aeration deficit (AF2) effecting photosyn.  (set to 1.0 for no stress run)'),
    FloatVar(
        'STEMN', 5, 2,
        info='0=original C to grain translocation, >0 to 1.0 sets % of C of stem to be transloc. to grain'
    ),
    FloatVar('MXNUP', 5, 2, info='max N uptake per day'),
    FloatVar('MXNCR', 5, 3, info='0.035=20%, .04=23% protein, max n:c ratio of grain growth'),
    FloatVar('WFNU', 5, 2, info='power term for water effect on N supply'),
    FloatVar('PNUPR', 5, 3, info='potential uptake rate (mg/meter/day)'),
    FloatVar('EXNO3', 5, 2, info='exponent for NO3 supply factor'),
    FloatVar('MNNO3', 5, 2, info='minimum for NO3 supply factor'),
    FloatVar('EXNH4', 5, 2, info='exponent for NH4 supply factor'),
    FloatVar('MNNH4', 5, 2, info='minimum for NH4 supply factor'),
    FloatVar('INGWT', 5, 2, info='initial grain weight  (mg/grain?)'),
    FloatVar(
        'INGNC', 5, 3,
        info='% protein, initial grain N conc   (init_grain_nconc or p_init_grain_nconc from APSIM Nwheat)'
    ),
    FloatVar(
        'FREAR', 5, 3,
        info='fraction between end ear and begin grainfilling, setting min stem weight for remobilisation'
    ),
    FloatVar('MNNCR', 5, 3, info='% protein, min n:c ratio of grain growth'),
    FloatVar(
        'GPPSS', 5, 2,
        info='gpp_start_stage - Grain per plant: 2=stem elong., 3=end leaf stage at which to start accumo. stem for '
             'gpp calc.'
    ),
    FloatVar(
        'GPPES', 5, 2,
        info='gpp_end_stage - Start grainfilling stage at which to end accumulation stem for Grain per plant calc'
    ),
    FloatVar('MXGWT', 5, 2, info='maximum kernal weight  [100 = no effect]'),
    FloatVar(
        'MNRTN', 5, 2,
        info='min root n due to grain n initialisation [0 = off] (root_n_min or p_root_n_min from APSIM Nwheat)'
    ),
    FloatVar('NOMOB', 5, 3, info='fraction of accum stem weight that is not mobile [0 = original]'),
    FloatVar('RTDP1', 5, 2, info='sw effect, 0 = old version'),
    FloatVar('RTDP2', 5, 2, info='crop stress effect, 0 = old version')
}

eco_vars_WHAPS = {
    CharacterVar('ECO#', 6, spc=0, info='Code for the ecotype to which a cultivar belongs (see *.cul file)'),
    CharacterVar('ECONAME', 17, header_fill='.', info='Name of the ecotype, which is referenced from *.CUL file'),

    FloatVar('TBASE', 5, 1, info='base temperature below which no development occurs, C'),
    FloatVar('TOPT', 5, 1, info='temperature at which maximum development rate occurs during vegetative stages, C'),
    FloatVar(
        'ROPT ', 5, 1, info='temperature at which maximum development rate occurs for reproductive stages, C (no '
                            'effect, WHAPS)'
    ),
    FloatVar('TTOP', 5, 1, info='temperature above which no development occurs, C'),
    FloatVar(
        'P20', 5, 1, info='Daylength below which daylength does not affect development rate, hours (no effect, WHAPS)'
    ),
    FloatVar('VREQ', 5, 0, info='Vernalization required for max.development rate (VDays)'),
    FloatVar('GDDE', 5, 1, info='Growing degree days per cm seed depth required for emergence, GDD/cm'),
    FloatVar('DSGFT', 5, 0, info='GDD from End Ear Growth to Start Grain Filling period, C'),
    FloatVar('RUE', 5, 1, info='Radiation use efficiency, g plant dry matter/MJ PAR (no effect, WHAPS)'),
    FloatVar('KCAN', 5, 2, info='Canopy light extinction coefficient for daily PAR. (no effect, WHAPS)'),
    FloatVar('TSEN', 5, 1, info='Critical temperature below which leaf damage occurs (default 6°C)'),
    FloatVar('CDAY', 5, 1, info='')
}



from tradssat.tmpl.var import CharacterVar, FloatVar
from ._utils import _return_vars

_standard_cul_vars = {
    'VAR#': {
        'class': CharacterVar,
        'args': dict(size=6, spc=0, info='Identification code or number for the specific cultivar.')
    },
    'VAR-NAME': {
        'class': CharacterVar,
        'args': dict(size=16, header_fill='.', info='Name of cultivar.')
    },
    'EXP#': {
        'class': CharacterVar,
        'args': dict(size=5, miss='.', info='Number of experiments,treatments used to estimate coefficients.')
    },
    'ECO#': {
        'class': CharacterVar,
        'args': dict(size=6, info='Code for the ecotype to which this cultivar belongs (see *.eco file)')
    },
    'G#WTS': {
        'class': FloatVar,
        'args': dict(size=5, dec=1, info='Standard grain number per unit canopy weight at anthesis (#/g)')
    },
    'GWTS': {
        'class': FloatVar,
        'args': dict(size=5, dec=0, info='Standard grain size,optimum conditions,normal plant density (mg)')
    },
    'LA1S': {
        'class': FloatVar,
        'args': dict(size=5, dec=1, info='Area of standard first leaf (cm2)')
    },
    'LAFV': {
        'class': FloatVar,
        'args': dict(size=5, dec=2, info='Increase in potential area of leaves,vegetative phase (fr/leaf)')
    },
    'LAFR': {
        'class': FloatVar,
        'args': dict(size=5, dec=2, info='Increase in potential area of leaves,reproductive phase (fr/leaf)')
    },
    'P1': {
        'class': FloatVar,
        'args': dict(size=5, dec=0, info='Duration of phase number 1 (PVoC.D)')
    },
    'P2': {
        'class': FloatVar,
        'args': dict(size=5, dec=0, info='Duration of phase number 2 (PVoC.D)')
    },
    'P3': {
        'class': FloatVar,
        'args': dict(
            size=5, dec=0,
            info='Duration of phase number 3 (PVoC.D) BUT P3 and P4 are \'dummy\' values inserted to set the total '
                 'duration of the P3+P4 phase. The actual balance between P3 and P4 is determined internally depending '
                 'on the leaf number at the end of phase 1.'
        )
    },
    'P4': {
        'class': FloatVar,
        'args': dict(
            size=5, dec=0,
            info='Duration of phase number 4 (PVoC.D) BUT P3 and P4 are \'dummy\' values inserted to set the total '
                 'duration of the P3+P4 phase. The actual balance between P3 and P4 is determined internally depending '
                 'othe leaf number at the end of phase 1.'
        )
    },
    'P5': {
        'class': FloatVar,
        'args': dict(size=5, dec=0, info='Duration of phase number 5 (PVoC.D)')
    },
    'P6': {
        'class': FloatVar,
        'args': dict(size=5, dec=0, info='Duration of phase number 6 (PVoC.D)')
    },
    'P7': {
        'class': FloatVar,
        'args': dict(size=5, dec=0, info='Duration of phase number 7 (PVoC.D)')
    },
    'P8': {
        'class': FloatVar,
        'args': dict(size=5, dec=1, info='Duration of phase number 8 (PVoC.D)')
    },
    'PPS1': {
        'class': FloatVar,
        'args': dict(size=5, dec=1, info='Photoperiod sensitivity,phase(1). % drop in rate,10h pp.change.')
    },
    'PPS2': {
        'class': FloatVar,
        'args': dict(size=5, dec=0, info='Photoperiod sensitivity,phase(2). % drop in rate,10h pp.change.')
    },
    'PHINT': {
        'class': FloatVar,
        'args': dict(size=5, dec=0, info='Interval between successive leaf appearances. (oC.d)')
    },
    'SHWTS': {
        'class': FloatVar,
        'args': dict(size=5, dec=1, info='Standard,non-stressed shoot dry weight (incl.grain),maturity (g)')
    },
    'SLAS': {
        'class': FloatVar,
        'args': dict(size=5, dec=0, info='Specific leaf area,standard (cm2/g)')
    },
    'VREQ': {
        'class': FloatVar,
        'args': dict(size=5, dec=4, info='Vernalization required for max.development rate (VDays)')
    },
    'VBASE': {
        'class': FloatVar,
        'args': dict(size=5, dec=0, info='Vernalization requirement before any effect (VDays)')
    },
    'VEFF': {
        'class': FloatVar,
        'args': dict(size=5, dec=2, info='Vernalization effect (Rate reduction when unvernalized (fr)')
    },

}
_standard_eco_vars = {
    'ECO#': {
        'class': CharacterVar,
        'args': dict(size=6, spc=0, info='Ecotype code (text)')
    },
    'ECONAME': {
        'class': CharacterVar,
        'args': dict(size=17, header_fill='.', info='Ecotype name')
    },
    'AWNS': {
        'class': FloatVar,
        'args': dict(size=5, dec=1, info='Awn score (0-10; 10=very long)')
    },
    'G#RF': {
        'class': FloatVar,
        'args': dict(size=5, dec=3, info='Grain number radiation factor (fr increase/(MJ/m2.d) radiation in phase 5)')
    },
    'GM%H': {
        'class': FloatVar,
        'args': dict(size=5, dec=1, info='Grain moisture percentage at harvest (%)')
    },
    'GN%MN': {
        'class': FloatVar,
        'args': dict(size=5, dec=1, info='Minimum grain nitrogen (%)')
    },
    'GN%S': {
        'class': FloatVar,
        'args': dict(size=5, dec=1, info='Standard grain nitrogen concentration (%)')
    },
    'GWTAF': {
        'class': FloatVar,
        'args': dict(size=5, dec=1, info='Grain weight adjustment factor (fr/10000kg/ha canopy wt>threshold)')
    },
    'GWTAT': {
        'class': FloatVar,
        'args': dict(size=5, dec=0, info='Grain weight adjustment threshold (kg/ha canopy weight at anthesis)')
    },
    'HTSTD': {
        'class': FloatVar,
        'args': dict(size=5, dec=0, info='Standard canopy height (cm)')
    },
    'KCAN': {
        'class': FloatVar,
        'args': dict(size=5, dec=2, info='PAR extinction coefficient (#)')
    },
    'LSENI': {
        'class': FloatVar,
        'args': dict(size=5, dec=1, info='Leaf senescence,intermediate phases (%/standard day)')
    },
    'LSPHE': {
        'class': FloatVar,
        'args': dict(size=5, dec=1, info='Leaf senescence (final) phase end stage (GrowthStage)')
    },
    'LSPHS': {
        'class': FloatVar,
        'args': dict(size=5, dec=1, info='Leaf senescence (final) phase start stage (GrowthStage)')
    },
    'NUPNF': {
        'class': FloatVar,
        'args': dict(size=5, dec=2, info='NO3 uptake vs conc exponent adjustment factor (0-2;0=no uptake,1=standard)')
    },
    'NUPWF': {
        'class': FloatVar,
        'args': dict(
            size=5, dec=1, info='Soil water effect on N uptake adjustment factor (0-2;1.0=no effect,1=standard)'
        )
    },
    'PARU2': {
        'class': FloatVar,
        'args': dict(size=5, dec=1, info='PAR conversion factor,after change (g dry matter/MJ)')
    },
    'PARUE': {
        'class': FloatVar,
        'args': dict(size=5, dec=1, info='PAR conversion factor,standard (g dry matter/MJ)')
    },
    'PHF3': {
        'class': FloatVar,
        'args': dict(size=5, dec=1, info='Factor by which PHINTS multiplied -> PHINT for particular phase (#)')
    },
    'PHL2': {
        'class': FloatVar,
        'args': dict(size=5, dec=1, info='Leaf # at end of phyllochron phase (ie,at which PHINT changes) (#)')
    },
    'RDGS': {
        'class': FloatVar,
        'args': dict(size=5, dec=1, info='Root depth growth rate,standard (cm/standard day)')
    },
    'RS%A': {
        'class': FloatVar,
        'args': dict(size=5, dec=0, info='Reserves concentration in tops at start of anthesis (%)')
    },
    'RTNUP': {
        'class': FloatVar,
        'args': dict(size=5, dec=3, info='NO3 uptake/root length  (mg N/cm.day)')
    },
    'SLAS': {
        'class': FloatVar,
        'args': dict(size=5, dec=0, info='Specific leaf area,standard (cm2/g)')
    },
    'SSPHE': {
        'class': FloatVar,
        'args': dict(size=5, dec=1, info='Stem senescence (final) phase end stage (GrowthStage)')
    },
    'SSPHS': {
        'class': FloatVar,
        'args': dict(size=5, dec=1, info='Stem senescence (final) phase start stage (GrowthStage)')
    },
    'TDFAC': {
        'class': FloatVar,
        'args': dict(size=5, dec=1, info='Tiller death factor (%/st.day when tiller wt 2xstandard wt)')
    },
    'TDPHE': {
        'class': FloatVar,
        'args': dict(size=5, dec=1, info='Tiller death phase end stage (GrowthStage)')
    },
    'TDPHS': {
        'class': FloatVar,
        'args': dict(size=5, dec=1, info='Tiller death phase start stage (GrowthStage)')
    },
    'TDSF': {
        'class': FloatVar,
        'args': dict(size=5, dec=2, info='Tiller death stress factor (#;0,2->no,full stress acceleration)')
    },
    'TIFAC': {
        'class': FloatVar,
        'args': dict(size=5, dec=1, info='Tiller initiation (rate) factor (fr of phyllochron based) (#)')
    },
    'TIL#S': {
        'class': FloatVar,
        'args': dict(size=5, dec=1, info='Tillering phase (production) start stage (leaf #)')
    },
    'TILPE': {
        'class': FloatVar,
        'args': dict(size=5, dec=1, info='Tillering phase (production) end (GrowthStage)')
    },
    'TKFH': {
        'class': FloatVar,
        'args': dict(size=5, dec=0, info='Cold tolerance when fully hardened (oC)')
    }
}


def cropsim_cul_vars(rename=None, exclude=None):
    return _return_vars(_standard_cul_vars, rename=rename, exclude=exclude)


def cropsim_eco_vars(rename=None, exclude=None):
    return _return_vars(_standard_eco_vars, rename=rename, exclude=exclude)



from tradssat.tmpl.output import OutFile
from tradssat.tmpl.var import FloatVar, IntegerVar


class SoilWatOut(OutFile):
    filename = 'SoilWat.Out'

    def _get_var_info(self):
        return vars_


vars_ = {
    IntegerVar('YEAR', 4, info='Year'),
    IntegerVar('DOY', 3, info='Day of year starting on Jan 1.'),
    IntegerVar('DAS', 3, info='Day after start'),
    IntegerVar('SWTD', 4, info='total soil water in profile, mm'),
    IntegerVar('SWXD', 4, info='potentially extractable water, cm'),
    IntegerVar('ROFC', 4, info='cumulative runoff'),
    IntegerVar('DRNC', 4, info='cumulative drainage'),
    IntegerVar('PREC', 4, info='cumulative precipitation'),
    IntegerVar('IR#C', 4, info='irrigation'),
    IntegerVar('IRRC', 4, info='cumulative irrigation'),
    IntegerVar('DTWT', 4, info='water table depth, cm'),
    FloatVar('MWTD', 4, 2, info='water stored in mulch layer, mm'),
    FloatVar('TDFD', 4, 0, info='water lost to tile drainage, mm/d'),
    FloatVar('TDFC', 4, 0, info='cumulative water lost to tile drainage, mm'),
    FloatVar('ROFD', 4, 0, info='total daily surface runoff, mm/d'),
    FloatVar('SW1D', 4, 3, info='soil water content at 0-5 cm depth, mm3/mm3'),
    FloatVar('SW2D', 4, 3, info='soil water content at 5-15 cm depth, mm3/mm3'),
    FloatVar('SW3D', 4, 3, info='soil water content at 15-30 cm depth, mm3/mm3'),
    FloatVar('SW4D', 4, 3, info='soil water content at 30-45 cm depth, mm3/mm3'),
    FloatVar('SW5D', 4, 3, info='soil water content at 45-60 cm depth, mm3/mm3'),
    FloatVar('SW6D', 4, 3, info='soil water content at 60-90 cm depth, mm3/mm3'),
    FloatVar('SW7D', 4, 3, info='soil water content at 90-110 cm depth, mm3/mm3'),
    FloatVar('SW8D', 4, 3, info='soil water content at 120-150 cm depth, mm3/mm3'),
    FloatVar('SW9D', 4, 3, info='soil water content at ____ cm depth, mm3/mm3'),
    FloatVar('SW10', 4, 3, info='soil water content at ____ cm depth, mm3/mm3'),

}



from tradssat.tmpl.output import FinalOutFile
from tradssat.tmpl.var import IntegerVar, CharacterVar, FloatVar


class SummaryOut(FinalOutFile):
    """
    Reader for DSSAT run summary (SUMMARY.OUT) output files.
    """
    filename = 'Summary.OUT'

    def _get_var_info(self):
        return _vars


_vars = {
    IntegerVar('RUNNO', 5, spc=4, info='Run number'),
    IntegerVar('TRNO', 5, spc=2, info='Run number'),
    IntegerVar('R#', 2, info='Repetition number'),
    IntegerVar('O#', 2, info=''),
    IntegerVar('C#', 2, info=''),
    CharacterVar('CR', 2, info='Crop code'),
    CharacterVar('MODEL', 8, header_fill='.', info='DSSAT crop model used'),
    CharacterVar('EXNAME', 8, header_fill='.', info='Experiment file name'),
    CharacterVar('TNAM', 25, header_fill='.', info='Treatment name'),
    CharacterVar('FNAM', 8, header_fill='.', info='Field name'),
    CharacterVar('WSTA', 8, header_fill='.', info='Weather station name'),
    IntegerVar('WYEAR', 4, info=''),
    CharacterVar('SOIL_ID', 10, header_fill='.', info='Soil name'),
    FloatVar('XLAT', 16, 9, info=''),
    FloatVar('LONG', 15, 9, info=''),
    IntegerVar('ELEV', 9, info=''),
    IntegerVar('SDAT', 7, info='Start date, YYYYDDD'),
    IntegerVar('PDAT', 7, info='Planting date, YYYYDDD'),
    IntegerVar('EDAT', 7, info='Emergence date, YYYYDDD'),
    IntegerVar('ADAT', 7, info='Anthesis date, start, YYYYDDD'),
    IntegerVar('MDAT', 7, info='Maturity (physiological) date, YYYYDDD'),
    IntegerVar('HDAT', 7, info='Harvest date, YYYYDDD'),
    IntegerVar('HYEAR', 7, info=''),
    IntegerVar('DWAP', 5, info='Planting material dry weight, kg ha-1'),
    IntegerVar('CWAM', 7, info='Canopy dry weight at maturity, kg ha-1'),
    IntegerVar('HWAM', 7, info='Yield at maturity, kg ha-1'),
    IntegerVar('HWAH', 7, info='Yield at harvest, kg ha-1'),
    IntegerVar('BWAH', 7, info='Harvested byproduct dry weight, kg ha-1'),
    IntegerVar('PWAM', 5, info=''),
    FloatVar('HWUM', 7, 3, info='Harvest product individual dry wt., mg or g '),
    IntegerVar('H#AM', 7, info='Harvest product number per m2 at maturity'),
    FloatVar('H#UM', 7, 1, info='Harvest product number per unit at maturity'),
    FloatVar('HIAM', 5, 3, info=''),
    FloatVar('LAIX', 5, 1, info=''),
    IntegerVar('FCWAM', 7, info=''),
    IntegerVar('FHWAM', 7, info=''),
    IntegerVar('HWAHF', 7, info=''),
    IntegerVar('FBWAH', 7, info=''),
    IntegerVar('FPWAM', 7, info=''),
    IntegerVar('IR#M', 5, info='Irrigation applications, number'),
    IntegerVar('IRCM', 5, info='Irrigation applied, cumulative, mm'),
    IntegerVar('PRCM', 5, info='Precipitation cumulative, mm'),
    IntegerVar('ETCM', 5, info='Evapotranspiration in season, cumulative, mm'),
    IntegerVar('EPCM', 5, info=''),
    IntegerVar('ESCM', 5, info=''),
    IntegerVar('ROCM', 5, info='Runoff, cumulative, mm'),
    IntegerVar('DRCM', 5, info='Cumulative drainage from profile, mm'),
    IntegerVar('SWXM', 5, info='Water remaining in soil, available, mm'),
    IntegerVar('NI#M', 5, info='Nitrogen applications, number'),
    IntegerVar('NICM', 5, info='Nitrogen applied, cumulative, kg ha-1'),
    IntegerVar('NFXM', 5, info='Nitrogen fixed, cumulative, kg ha-1'),
    IntegerVar('NUCM', 5, info='Total N uptake during season, kg/ha'),
    IntegerVar('NLCM', 5, info='Nitrogen leached, cumulative, kg ha-1'),
    IntegerVar('NIAM', 5, info='Inorganic N in soil at maturity, kg/ha'),
    IntegerVar('CNAM', 5, info='Tops N at maturity, kg ha-1 '),
    IntegerVar('GNAM', 5, info='Nitrogen in harvest product (seed), kg ha-1 '),
    FloatVar('N2OEC', 5, 3, info=''),
    IntegerVar('PI#M', 5, info=''),
    IntegerVar('PICM', 5, info=''),
    IntegerVar('PUPC', 5, info=''),
    IntegerVar('SPAM', 5, info='Cumulative plant P uptake, kg/ha'),
    IntegerVar('KI#M', 5, info=''),
    IntegerVar('KICM', 5, info=''),
    IntegerVar('KUPC', 5, info=''),
    IntegerVar('SKAM', 5, info=''),
    IntegerVar('RECM', 5, info='Residue, etc. applied, cumulative, kg ha-1'),
    IntegerVar('ONTAM', 6, info=''),
    IntegerVar('ONAM', 6, info='Soil organic N at end of season, kg ha-1'),
    IntegerVar('OPTAM', 6, info=''),
    IntegerVar('OPAM', 6, info=''),
    IntegerVar('OCTAM', 7, info=''),
    IntegerVar('OCAM', 7, info='Soil C at end of season, t ha-1'),
    IntegerVar('CO2EC', 7, info=''),
    FloatVar('CH4EC', 6, 1, info=''),
    FloatVar('DMPPM', 8, 1, info=''),
    FloatVar('DMPEM', 8, 1, info=''),
    FloatVar('DMPTM', 8, 1, info=''),
    FloatVar('DMPIM', 8, 1, info=''),
    FloatVar('YPPM', 8, 1, info=''),
    FloatVar('YPEM', 8, 1, info=''),
    FloatVar('YPTM', 8, 1, info=''),
    FloatVar('YPIM', 8, 1, info=''),
    FloatVar('DPNAM', 8, 1, info=''),
    FloatVar('DPNUM', 8, 1, info=''),
    FloatVar('YPNAM', 8, 1, info=''),
    FloatVar('YPNUM', 8, 1, info=''),
    IntegerVar('NDCH', 5, info=''),
    FloatVar('TMAXA', 5, 1, info=''),
    FloatVar('TMINA', 5, 1, info=''),
    FloatVar('SRADA', 5, 1, info=''),
    FloatVar('DAYLA', 5, 1, info=''),
    FloatVar('CO2A', 6, 1, info=''),
    FloatVar('PRCP', 6, 1, info=''),
    FloatVar('ETCP', 6, 1, info=''),
    FloatVar('ESCP', 6, 1, info=''),
    FloatVar('EPCP', 6, 1, info=''),

    # Recently added (#19)
    IntegerVar('P#', 2, info=''),
    IntegerVar('NMINC', 5, info=''),
}



from tradssat.tmpl import InpFile
from .exper_vars import main_vars, header_vars


class ExpFile(InpFile):
    """
    File reader for all DSSAT (.ccX) experiment files.
    """
    ext = [
        '.ALX', '.ARX', '.BAX', '.BNX', '.BWX', '.COX', '.CSX', '.FAX', '.GWX', '.MLX', '.MZX', '.PIX', '.PNX', '.PTX',
        '.RIX', '.SBX', '.SCX', '.SGX', '.STX', '.WHX', '.SQX', '.SNX'
    ]  # Todo: add all DSSAT 47 crop extension codes

    def _get_var_info(self):
        return main_vars

    def _get_header_vars(self):
        return {'EXP.DETAILS: ': header_vars}



from tradssat.tmpl.var import CharacterVar, FloatVar

cul_vars_BACER = {
    CharacterVar('VAR#', 6, spc=0, info='Identification code or number for the specific cultivar.'),
    CharacterVar('VAR-NAME', 16, header_fill='.', info='Name of cultivar.'),
    CharacterVar('EXP#', 5, miss='.', info='Number of experiments used to generate parameters'),
    CharacterVar('ECO#', 6, info='Ecotype code for this cultivar,points to entry in ECO file'),
    FloatVar('P1V', 5, 0, info='Days,optimum vernalizing temperature,required for vernalization'),
    FloatVar('P1D', 5, 0, info='Photoperiod response (% reduction in rate/10 h drop in pp)'),
    FloatVar('P5', 5, 0, info='Grain filling (excluding lag) phase duration (oC.d)'),
    FloatVar('G1', 5, 0, info='Kernel number per unit canopy weight at anthesis (#/g)'),
    FloatVar('G2', 5, 0, info='Standard kernel size under optimum conditions (mg)'),
    FloatVar('G3', 5, 1, info='Standard,non-stressed mature tiller wt (incl grain) (g dwt)'),
    FloatVar('PHINT', 5, 0, info='Interval between successive leaf tip appearances (oC.d)'),
}

eco_vars_BACER = {
    CharacterVar('ECO#', 6, spc=0, info='Code for the ecotype (code)'),
    FloatVar('GN%MN', 5, 1, info='Minimum grain N (%)'),
    FloatVar('GN%S', 5, 1, info='Standard grain N (%)'),
    FloatVar('HTSTD', 5, 0, info='Standard canopy height (cm)'),
    FloatVar('KCAN', 5, 2, info='PAR extinction coefficient (#)'),
    FloatVar('LA1S', 5, 1, info='Area of standard first leaf (cm2)'),
    FloatVar('LAFR', 5, 2, info='Increase in potential area of leaves,reproductive phase (fr/leaf)'),
    FloatVar('LAFV', 5, 2, info='Increase in potential area of leaves,vegetative phase (fr/leaf)'),
    FloatVar('LARS', 5, 1, info='Area of standard reproductive phase leaf (cm2)'),
    FloatVar('LAVS', 5, 1, info='Area of standard vegetative phase leaf (cm2)'),
    FloatVar('LLIFE', 5, 2, info='Life of leaves during vegetative phase (phyllochrons)'),
    FloatVar('LSPHE', 5, 1, info='Final leaf senescence ends (GrowhStage)'),
    FloatVar('LSPHS', 5, 1, info='Final leaf senescence starts (GrowhStage)'),
    FloatVar('P1', 5, 0, info='Duration of phase end juvenile to terminal spikelet (PVTU)'),
    FloatVar('P2', 5, 0, info='Duration of phase terminal spikelet to end leaf growth (TU)'),
    FloatVar('P2FR1', 5, 2, info='Duration of phase terminal spikelet to jointing (fr P2)'),
    FloatVar('P3', 5, 0, info='Duration of phase end leaf growth to end spike growth (TU)'),
    FloatVar('P4', 5, 0, info='Duration of phase end spike growth to end grain fill lag (TU)'),
    FloatVar('P4FR1', 5, 2, info='Duration of phase end spike growth to anthesis (fr P4)'),
    FloatVar('P4FR2', 5, 2, info='Duration of phase anthesis start to anthesis end (fr P4)'),
    FloatVar('PARU2', 5, 1, info='PAR conversion to dm ratio,after last leaf (g/MJ) (If -99,set to PARUE)'),
    FloatVar('PARUE', 5, 1, info='PAR conversion to dm ratio,before last leaf stage (g/MJ)'),
    FloatVar('PHF3', 5, 2, info='Factor by which PHINTS is multiplied -> PHINT for phase (#)'),
    FloatVar('PHL2', 5, 0, info='Leaf # produced during phyllochron phase (#)'),
    FloatVar('PPFPE', 5, 3, info='Daylength factor,pre emergence (#,0-1)'),
    FloatVar('RDGS', 5, 1, info='Root depth growth rate,early phase (cm/standard d)'),
    FloatVar('RS%S', 5, 0, info='Reserves part of assimilates going to stem (%)'),
    FloatVar('SLAS', 5, 0, info='Specific leaf area,standard first leaf (cm2/g)'),
    FloatVar('TDFAC', 5, 1, info='Tiller death factor (%/st.day when tiller wt 2xstandard wt)'),
    FloatVar('TDPHE', 5, 1, info='Tiller death ending stage (GrowthStage)'),
    FloatVar('TDPHS', 5, 1, info='Tiller death start stage (GrowthStage)'),
    FloatVar('TIFAC', 5, 1, info='Tiller initiation (rate) factor (fr of phyllochron based) (#)'),
    FloatVar('TIL#S', 5, 1, info='Tiller production starts (leaf #)'),
    FloatVar('TIPHE', 5, 1, info='Tillering phase end stage (GrowthStage)'),
    FloatVar('TKFH', 5, 0, info='Temperature at which killed when fully hardened (oC)'),
    FloatVar('VEFF', 5, 1, info='Vernalization effect (Rate reduction when unvernalized (fr)'),
    FloatVar('WFGU', 5, 2, info='Water stress factor,growth,upper (fr)'),
    FloatVar('WFPU', 5, 2, info='Water stress factor,photosynthesis,upper (fr)'),
    FloatVar('AWNS', 5, 2, info='Awn score (0-10;10=very long)')

}



from tradssat.tmpl.output import OutFile
from tradssat.tmpl.var import FloatVar, IntegerVar


class MulchOut(OutFile):
    filename = 'Mulch.Out'

    def _get_var_info(self):
        return vars_


vars_ = {
    IntegerVar('YEAR', 4, info='Year'),
    IntegerVar('DOY', 3, info='Day of year starting on Jan 1.'),
    IntegerVar('DAS', 5, info='Day after start'),
    FloatVar('MCFD', 4, 3, info='fraction of soil covered by mulch'),
    FloatVar('MDEPD', 5, 2, info='thickness of mulch layer, cm'),
    IntegerVar('MWAD', 4, info='mass of mulch layer, kg/ha'),
    FloatVar('MWTD', 4, 2, info='water stored in mulch layer, mm'),
}



from tradssat.tmpl.var import CharacterVar, FloatVar

cul_vars_TRARO = {
    CharacterVar('VAR#', 6, spc=0, info='Identification code or number for the specific cultivar.'),
    CharacterVar('VAR-NAME', 16, header_fill='.', info='Name of cultivar.'),
    CharacterVar('EXPNO', 5, miss='.', info='Number of experiments used for calibration.'),
    CharacterVar('ECO#', 6, info='Code for the ecotype to which this cultivar belongs (see *.eco file)'),

    FloatVar('P1', 5, 0),
    FloatVar('P3', 5, 0),
    FloatVar('P4', 5, 0),
    FloatVar('P5', 5, 0),
    FloatVar('G2', 5, 2),
    FloatVar('G3', 5, 2),
    FloatVar('G4', 5, 2),
    FloatVar('PHINT', 5, 1),
    FloatVar('PCINT', 5, 1),
    FloatVar('PCGRD', 5, 2)
}



from tradssat.tmpl.var import FloatVar, CharacterVar

cul_vars_PIALO = {
    CharacterVar('VAR#', 6, spc=0, info=''),
    CharacterVar('VAR-NAME', 16, header_fill='.', info=''),
    CharacterVar('EXPNO', 5, miss='', info=''),
    CharacterVar('ECO#', 6, info=''),

    FloatVar('P1', 5, 1, info='Growing degree days from first leaf emerged to end of stem growth'),
    FloatVar('P2', 5, 0, info='Growing degree days from forcing to sepals closed on youngest flower'),
    FloatVar('P3', 5, 0, info='Growing degree days from SCY to early flowering'),
    FloatVar('P4', 5, 0, info='Cumulative growing degree days from early flowering to maturity'),
    FloatVar('P5', 5, 0, info='Growing degree days from fruit harvest to physiological maturity'),
    FloatVar('P6', 5, 1, info='Growing degree days from root initiation to first leaf emerged'),
    FloatVar('G2', 5, 0, info='Potential eye number'),
    FloatVar('G3', 5, 1, info='Potential eye growth rate'),
    FloatVar('PHINT', 5, 1, info='')
}



import os

import numpy as np

from tradssat import CULFile, ECOFile
from .mgr import get_dssat_subdir, PeriphFileMgr


class PeriphGenMgr(PeriphFileMgr):

    def __init__(self, crops, cultivars, levels, model=None):
        self.crops = crops
        self.cultivars = cultivars

        self.files = {
            lvl: GeneticMgr(_get_model(crp, model), cult) for lvl, crp, cult in zip(levels, crops, cultivars)
        }

    def get_value(self, var, level):
        return self.files[level].get_value(var)

    def set_value(self, var, val, level):
        return self.files[level].set_value(var, val)

    def variables(self):
        return {str(vr) for f in self.files.values() for vr in f.variables()}


class GeneticMgr(object):

    def __init__(self, crop, cult):
        self.crop = crop
        self.cult = cult

        geno_dir = get_dssat_subdir('Genotype')

        self.eco_file = self.cult_file = None
        for f in os.listdir(geno_dir):
            name = os.path.split(f)[1]
            if CULFile.matches_file(name) and name.startswith(self.crop):
                self.cult_file = CULFile(os.path.join(geno_dir, f))
            elif ECOFile.matches_file(name) and name.startswith(self.crop):
                self.eco_file = ECOFile(os.path.join(geno_dir, f))

            if self.eco_file is not None and self.cult_file is not None:
                break

        if self.cult_file is None:
            raise ValueError('No cultivar file (.CUL) found for crop "{}".'.format(self.crop))

        eco_codes = self.cult_file.get_value('ECO#')
        cult_codes = self.cult_file.get_value('VAR#')
        if self.cult not in cult_codes:
            raise ValueError('Cultivar {} not found in cultivar file (.CUL).'.format(self.cult))
        eco = eco_codes[cult_codes == self.cult]

        self.eco_n = np.where(self.eco_file.get_value('ECO#') == eco)
        self.cult_n = np.where(cult_codes == cult)

    def get_value(self, var):
        if var in self.cult_file.variables():
            return self.cult_file.get_value(var, cond={'VAR#': self.cult})
        elif self.eco_file is not None and var in self.eco_file.variables():
            return self.eco_file.get_value(var)
        else:
            raise ValueError('No genetic variable named "{}" was found.'.format(var))

    def set_value(self, var, val):
        if var in self.cult_file.variables():
            return self.cult_file.set_value(var, val)
        elif self.eco_file is not None and var in self.eco_file.variables():
            return self.eco_file.set_value(var, val)
        else:
            raise ValueError('No genetic variable named "{}" was found.'.format(var))

    def variables(self):
        vars_ = {*self.cult_file.variables()}
        if self.eco_file is not None:
            vars_.update(self.eco_file.variables())

        return vars_


_crop_models = {
    'AL': ['ALFRM'],
    'BA': ['BACER', 'BACRP'],
    'BH': ['BHGRO'],
    'BM': ['BMFRM', 'BMGRO'],
    'BN': ['BNGRO'],
    'BR': ['BRFRM', 'BRGRO'],
    'BS': ['BSCER'],
    'CB': ['CBGRO'],
    'CH': ['CHGRO'],
    'CN': ['CNGRO'],
    'CO': ['COGRO'],
    'CP': ['CPGRO'],
    'CS': ['CSCAS', 'CSYCA'],
    'FB': ['FBGRO'],
    'G0': ['G0GRO'],
    'GB': ['GBGRO'],
    'ML': ['MLCER'],
    'MZ': ['MZCER', 'MZIXM'],
    'PI': ['PIALO'],
    'PN': ['PNGRO'],
    'PP': ['PPGRO'],
    'PR': ['PRGRO'],
    'PT': ['PTSUB'],
    'RI': ['RICER', 'RIORZ'],
    'SB': ['SBGRO'],
    'SC': ['SCCAN', 'SCCSP'],
    'SF': ['SFGRO'],
    'SG': ['SGCER'],
    'SU': ['SUGRO'],
    'SW': ['SWCER'],
    'TM': ['TMGRO'],
    'TN': ['TNARO'],
    'VB': ['VBGRO'],
    'WH': ['WHAPS', 'WHCER', 'WHCRP']
}


def _get_model(crop, mod):
    try:
        m = _crop_models[crop]
    except KeyError:
        raise ValueError('No model defined for crop "{}" in traDSSAT.'.format(crop))
    if mod in m:
        return mod
    else:
        return m[0]



import os

from chardet import UniversalDetector


def read_txt(f):
    with open(f, 'r', encoding='UTF-8') as d:
        return d.read()


def detect_encod(file):
    detector = UniversalDetector()
    with open(file, 'rb') as d:
        for line in d.readlines():

            detector.feed(line)

            if detector.done:
                break

    detector.close()

    return detector.result['encoding']


DSSAT_DIR = 'DSSAT_DIR'
config = {DSSAT_DIR: None}


def set_dssat_dir(dssat_dir):
    if os.path.isdir(dssat_dir):
        config[DSSAT_DIR] = dssat_dir
    else:
        raise FileNotFoundError(dssat_dir)


def get_dssat_dir():
    return config[DSSAT_DIR]



from tradssat.tmpl.var import CharacterVar, FloatVar

cul_vars_SGCER = {
    CharacterVar('VAR#', 6, spc=0, info='Identification code or number for the specific cultivar.'),
    CharacterVar('VAR-NAME', 16, header_fill='.', info='Name of cultivar.'),
    CharacterVar('EXPNO', 5, miss='.', info='Number of experiments used for calibration.'),
    CharacterVar('ECO#', 6, info='Code for the ecotype to which this cultivar belongs (see *.eco file)'),

    FloatVar(
        'P1', 5, 1,
        info='Thermal time from seedling emergence to the end of the juvenile phase (expressed in degree days above '
             'TBASE during which the plant is not responsive to changes in photoperiod'
    ),
    FloatVar(
        'P2', 5, 1,
        info='Thermal time from the end of the juvenile stage to tassel initiation under short days (degree days above '
             'TBASE)'
    ),
    FloatVar(
        'P2O', 5, 2,
        info='Critical photoperiod or the longest day length (in hours) at which development occurs at a maximum rate.'
             ' At values higher than P2O, the rate of development is reduced'
    ),
    FloatVar(
        'P2R', 5, 1,
        info='Extent to which phasic development leading to panicle initiation (expressed in degree days) is delayed '
             'for each hour increase in photoperiod above P2O'
    ),
    FloatVar('PANTH', 5, 1,
             info='Thermal time from the end of tassel initiation to anthesis (degree days above TBASE)'),
    FloatVar('P3', 5, 1, info='Thermal time from to end of flag leaf expansion to anthesis (degree days above TBASE)'),
    FloatVar('P4', 5, 1, info='Thermal time from anthesis to beginning grain filling (degree days above TBASE)'),
    FloatVar('P5', 5, 1, info='Thermal time from beginning of grain filling to physiological maturity (degree days '
                              'above TBASE)'),
    FloatVar(
        'PHINT', 5, 2,
        info='Phylochron interval; the interval in thermal time between successive leaf tip appearances (degree days)'
    ),
    FloatVar('G1', 5, 1, info='Scaler for relative leaf size'),
    FloatVar('G2', 5, 1, info='Scaler for partitioning of assimilates to the panicle (head).'),
    FloatVar('PSAT', 5, 1, info='Critical photoperiod below which development is not delayed (optional)'),
    FloatVar('PBASE', 5, 1, info='Ceiling photoperiod above which development is delayed indefinitely (optional)')
}

eco_vars_SGCER = {
    CharacterVar('ECO#', 6, spc=0, info='Code for the ecotype to which a cultivar belongs (see *.cul file)'),
    CharacterVar('ECONAME', 17, header_fill='.', info='Name of the ecotype, which is referenced from *.CUL file'),

    FloatVar('TBASE', 5, 1, info='Base temperature below which no development occurs (oC)'),
    FloatVar('TOPT', 5, 1, info='Temperature at which maximum development occurs for vegetative stages (oC)'),
    FloatVar('ROPT', 5, 1, info='Temperature at which maximum development occurs for reproductive stages (oC)'),
    FloatVar('GDDE', 5, 1, info='Growing degree days per cm seed depth required for emergence (degree days/cm)'),
    FloatVar('RUE', 5, 1, info='Radiation use efficiency (g plant dry matter/MJ PAR)'),
    FloatVar('KCAN', 5, 2, info='Canopy light extinction coefficient for daily PAR'),
    FloatVar('STPC', 5, 3, info='Partitioning to stem growth as a fraction of potential leaf growth'),
    FloatVar('RTPC', 5, 3, info='Partitioning to root growth as a fraction of available carbohydrates'),
    FloatVar('TILFC', 5, 1, info='Tillering factor (0.0 no tillering; 1.0 full tillering)'),
    FloatVar('PLAM', 5, 1, info='Plant leaf area maximun (Initial leaf area)')
}
